@@ -26,11 +26,13 @@
 #include <arpa/inet.h>		/* BSD 4.3: inet_aton */
 #include <dirent.h>		/* ? */
 #include <inttypes.h>		/* POSIX.1 (CX): PRI* */
+#include "iconvme.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #include "quotearg.h"		/* Gnulib */
+#include "human.h"		/* Gnulib */
 #include "fnmatch.h"		/* Gnulib/GNU Libc */
 #include "quote.h"		/* Gnulib */
 #include "stpcpy.h"		/* Gnulib/GNU Libc */
@@ -77,6 +79,7 @@ static void cmd_unsearch(int argc, char **argv);
 static void cmd_alias(int argc, char **argv);
 static void cmd_unalias(int argc, char **argv);
 static void cmd_shell(int argc, char **argv);
+static void cmd_lookup(int argc, char **argv);
 static void shell_command_completion_selector(DCCompletionInfo *ci);
 static void executable_completion_generator(DCCompletionInfo *ci);
 static void alias_completion_generator(DCCompletionInfo *ci);
@@ -298,6 +301,9 @@ command_init(void)
           "shell will be started (SHELL environment variable or `/bin/sh' if that "
           "is not set). microdc will continue in the background while the command "
           "is executing.\n"));
+    add_builtin_command("lookup", cmd_lookup, NULL,
+        _("lookup HOST ..."),
+        _("Lookup the IP address of specified hosts.\n"));
 
     add_alias_command("ll", "ls -l");
 }
@@ -698,13 +704,17 @@ cmd_status(int argc, char **argv)
 {
     HMapIterator it;
     uint32_t c;
+    char sizebuf[LONGEST_HUMAN_READABLE+1];
 
     switch (hub_state) {
     case DC_HUB_DISCONNECTED:
 	screen_putf(_("Hub state: %s\n"), _("Not connected"));
 	break;
+    case DC_HUB_LOOKUP:
+        screen_putf(_("Hub state: %s\n"), _("Looking up IP address"));
+        break;
     case DC_HUB_CONNECT:
-	screen_putf(_("Hub state %s\n"), _("Waiting for complete connection"));
+	screen_putf(_("Hub state: %s\n"), _("Waiting for complete connection"));
 	break;
     case DC_HUB_LOCK:
 	screen_putf(_("Hub state: %s\n"), _("Waiting for $Lock"));
@@ -730,18 +740,18 @@ cmd_status(int argc, char **argv)
     if (c == 0)
         screen_putf(_("  (none)\n"));
 
-    screen_putf(_("Share size: %" PRIu64" %s (%" PRIu64 " MB)\n"),
+    screen_putf(_("Share size: %" PRIu64" %s (%s)\n"),
                 my_share_size,
                 ngettext("byte", "bytes", my_share_size),
-                my_share_size/(1024*1024));
-    screen_putf(_("Bytes received: %" PRIu64 " %s (%" PRIu64 " MB)\n"),
+                human_readable(my_share_size, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
+    screen_putf(_("Bytes received: %" PRIu64 " %s (%s)\n"),
                 bytes_received,
                 ngettext("byte", "bytes", bytes_received),
-                bytes_received/(1024*1024));
-    screen_putf(_("Bytes sent: %" PRIu64 " %s (%" PRIu64 " MB)\n"),
+                human_readable(bytes_received, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
+    screen_putf(_("Bytes sent: %" PRIu64 " %s (%s)\n"),
                 bytes_sent,
                 ngettext("byte", "bytes", bytes_sent),
-                bytes_sent/(1024*1024));
+                human_readable(bytes_sent, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
 }
 
 static void
@@ -757,8 +767,8 @@ cmd_exit(int argc, char **argv)
 static void
 cmd_say(int argc, char **argv)
 {
-    char *msg;
-    char *tmp;
+    char *t1;
+    char *t2;
 
     if (argc <= 1) {
     	screen_putf(_("Usage: %s MESSAGE..\n"), argv[0]);
@@ -768,19 +778,24 @@ cmd_say(int argc, char **argv)
     	screen_putf(_("Not connected.\n"));
 	return;
     }
-    tmp = join_strings(argv+1, argc-1, ' ');
-    msg = escape_message(tmp);
+    t1 = join_strings(argv+1, argc-1, ' ');
+    if (to_hub_iconv != (iconv_t) -1) {
+        t2 = iconv_alloc(to_hub_iconv, t1);
+        free(t1);
+        t1 = t2;
+    }
+    t2 = escape_message(t1);
+    free(t1);
     /* Don't print on screen - hub will send us the message back. */
-    hub_putf("<%s> %s|", my_nick, msg); /* Ignore error */
-    free(msg);
-    free(tmp);
+    hub_putf("<%s> %s|", my_nick, t2); /* Ignore error */
+    free(t2);
 }
 
 static void
 cmd_msg(int argc, char **argv)
 {
-    char *msg;
-    char *tmp;
+    char *t1;
+    char *t2;
     DCUserInfo *ui;
 
     if (argc <= 2) {
@@ -797,12 +812,18 @@ cmd_msg(int argc, char **argv)
 	return;
     }
 
-    tmp = join_strings(argv+2, argc-2, ' ');
-    msg = escape_message(tmp);
-    screen_putf("Private to %s: <%s> %s\n", quotearg_n(0, ui->nick), quotearg_n(1, my_nick), tmp);
-    hub_putf("$To: %s From: %s $<%s> %s|", ui->nick, my_nick, my_nick, msg); /* Ignore error */
-    free(msg);
-    free(tmp);
+
+    t1 = join_strings(argv+2, argc-2, ' ');
+    screen_putf("Private to %s: <%s> %s\n", quotearg_n(0, ui->nick), quotearg_n(1, my_nick), t1); /* XXX: quotearg_n(2, t1)? */
+    if (to_hub_iconv != (iconv_t) -1) {
+        t2 = iconv_alloc(to_hub_iconv, t1);
+        free(t1);
+        t1 = t2;
+    }
+    t2 = escape_message(t1);
+    free(t1);
+    hub_putf("$To: %s From: %s $<%s> %s|", ui->nick, my_nick, my_nick, t2); /* Ignore error */
+    free(t2);
 }
 
 static void
@@ -828,44 +849,29 @@ static void
 cmd_connect(int argc, char **argv)
 {
     char *portstr;
-    struct sockaddr_in addr;
+    uint16_t port;
 
     if (argc == 1) {
     	screen_putf(_("Usage: %s HOST[:PORT]\n"), argv[0]);
 	return;
     }
     if (hub_state != DC_HUB_DISCONNECTED) {
-    	screen_putf(_("Already connected or connecting, disconnect first.\n"));
+    	screen_putf(_("Connection in progress, disconnect first.\n"));
 	return;
     }
 
     portstr = strchr(argv[1], ':');
     if (portstr != NULL) {
     	*portstr = '\0';
-    	if (!parse_uint16(portstr+1, &addr.sin_port)) {
-	    screen_putf(_("Invalid port number %s\n"), quote(portstr));
+    	if (!parse_uint16(portstr+1, &port)) {
+	    screen_putf(_("Invalid port number %s\n"), quote(portstr+1));
 	    return;
 	}
-	addr.sin_port = htons(addr.sin_port);
     } else {
-    	addr.sin_port = htons(DC_HUB_TCP_PORT);
+        port = DC_HUB_TCP_PORT;
     }
 
-    if (!inet_aton(argv[1], &addr.sin_addr)) {
-    	struct hostent *he;
-
-	screen_putf(_("Looking up IP address for %s\n"), quotearg(argv[1]));
-	he = gethostbyname(argv[1]);
-	if (he == NULL) {
-    	    screen_putf(_("%s: Cannot look up address - %s\n"), quotearg(argv[1]), hstrerror(h_errno));
-	    return;
-	}
-    	addr.sin_addr = *(struct in_addr *) he->h_addr;
-    }
-
-    addr.sin_family = AF_INET;
-
-    hub_connect(&addr); /* Ignore errors */
+    hub_new(argv[1], port);
 }
 
 static void
@@ -944,6 +950,32 @@ queued_file_cmp(const char *filename, DCQueuedFile *qf)
     return strcmp(filename, qf->filename);
 }
 
+void
+browse_list_parsed(DCFileList *node, void *data)
+{
+    char *nick = data;
+
+    /* These checks are to protect from the case when
+     * the user is already browsing the user the file
+     * list was parsed for.
+     */
+    if (browse_list == NULL && browse_user != NULL
+            && strcmp(nick, browse_user->nick) == 0) {
+	browse_list = node;
+	browse_path = xstrdup("/");
+	browse_path_previous = NULL;
+	update_prompt();
+	/* browse_list_parsed will never be called when browsing ourselves,
+	 * because our filelist is already parsed and always available.
+	 */
+	screen_putf(_("Now browsing %s.\n"), quotearg(browse_user->nick));
+    } else {
+	filelist_free(node);
+    }
+
+    free(nick);
+}
+
 /* XXX: move queue.c/browse.c? */
 static void
 cmd_browse(int argc, char **argv)
@@ -1012,14 +1044,13 @@ cmd_browse(int argc, char **argv)
     }
 
     browse_none();
-    browse_list = filelist_open(filename);
-    browse_path = xstrdup("/");
-    browse_path_previous = NULL;
     browse_user = ui;
     browsing_myself = false;
     ui->refcount++;
-    free(filename);
     update_prompt();
+
+    add_parse_request(browse_list_parsed, filename, xstrdup(ui->nick));
+    free(filename);
 }
 
 /* XXX: move queue.c/browse.c? */
@@ -1788,17 +1819,17 @@ append_download_file(DCUserInfo *ui, DCFileList *node, DCFileList *basenode, uin
             queued->base_path = filelist_get_path_with_trailing_slash(basenode);
             queued->flag = DC_TF_NORMAL;
             queued->status = DC_QS_QUEUED;
-            queued->length = node->u.reg.size;
+            queued->length = node->size;
             ptrv_append(ui->download_queue, queued);
         }
 
-        (*byte_count) += node->u.reg.size;
+        (*byte_count) += node->size;
         (*file_count)++;
     }
     else if (node->type == DC_TYPE_DIR) {
         HMapIterator it;
 
-        hmap_iterator(node->u.dir.children, &it);
+        hmap_iterator(node->dir.children, &it);
         while (it.has_next(&it))
             append_download_file(ui, it.next(&it), basenode, file_count, byte_count);
     }
@@ -1880,3 +1911,38 @@ cmd_get(int argc, char **argv)
         }
     }
 }
+
+static void
+lookup_address_looked_up(int rc, struct addrinfo *ai, void *data)
+{
+    char *host = data;
+
+    screen_putf("%s:", quotearg(host));
+    if (rc == 0) {
+        for (; ai != NULL; ai = ai->ai_next) {
+            struct sockaddr_in *addr = (struct sockaddr_in *) ai->ai_addr;
+            screen_putf(" %s", inet_ntoa(addr->sin_addr));
+        }
+    } else {
+        screen_putf(" %s", gai_strerror(rc));
+    }
+    screen_putf("\n");
+
+    free(host);
+}
+
+static void
+cmd_lookup(int argc, char **argv)
+{
+    int c;
+
+    if (argc == 1) {
+        screen_putf(_("missing host argument\n"));
+        return;
+    }
+
+    for (c = 1; c < argc; c++) {
+        struct addrinfo hints = { 0, PF_INET, SOCK_STREAM, 0, };
+        add_lookup_request(argv[c], NULL, &hints, lookup_address_looked_up, xstrdup(argv[c]));
+    }
+}
