@@ -77,7 +77,7 @@ fs_completion_entry_compare(const void *e1, const void *e2)
     return ce1->sorting.file_type - ce2->sorting.file_type;
 }   
 
-static DCFileList *
+DCFileList *
 new_file_node(const char *name, DCFileType type, DCFileList *parent)
 {
     DCFileList *node;
@@ -87,16 +87,15 @@ new_file_node(const char *name, DCFileType type, DCFileList *parent)
     node->type = type;
     node->parent = parent;
     if (parent != NULL)
-        hmap_put(parent->u.dir.children, node->name, node);
+        hmap_put(parent->dir.children, node->name, node);
+    node->size = 0;
     switch (node->type) {
     case DC_TYPE_DIR:
-    	node->u.dir.children = hmap_new();
-	node->u.dir.totalsize = 0;
+    	node->dir.children = hmap_new();
 	break;
     case DC_TYPE_REG:
-    	node->u.reg.size = 0;
-    default:
-    	break;
+        /* No operation at the moment! */
+        break;
     }
 
     return node;
@@ -109,7 +108,7 @@ get_child_node(DCFileList *node, const char *path)
         return node;
     if (IS_PARENT_DIR(path))
         return node->parent == NULL ? node : node->parent;
-    return hmap_get(node->u.dir.children, path);
+    return hmap_get(node->dir.children, path);
 }
 
 void
@@ -119,85 +118,14 @@ filelist_free(DCFileList *node)
     case DC_TYPE_REG:
     	break;
     case DC_TYPE_DIR:
-        hmap_foreach_value(node->u.dir.children, filelist_free);
-	hmap_free(node->u.dir.children);
+        hmap_foreach_value(node->dir.children, filelist_free);
+	hmap_free(node->dir.children);
     	break;
     }
     free(node->name);
     free(node);
 }
 
-DCFileList *
-parse_decoded_dclst(char *decoded, uint32_t decoded_len)
-{
-    DCFileList *node;
-    uint32_t c;
-    PtrV *dirs;
-
-    dirs = ptrv_new();
-    node = new_file_node("", DC_TYPE_DIR, NULL);
-    ptrv_append(dirs, node);
-    for (c = 0; c < decoded_len; c++) {
-    	DCFileList *oldnode;
-	char *name;
-    	int depth;
-
-	for (; c < decoded_len && decoded[c] == '\n'; c++);
-    	depth = 1;
-	for (; c < decoded_len && decoded[c] == '\t'; c++)
-	    depth++;
-	if (c >= decoded_len)
-	    break; /* Premature end */
-	if (decoded[c] == '\r') {
-	    c++; /* skip LF */
-	    continue; /* Skipping bad line */
-	}
-
-    	name = decoded + c;
-    	for (; c < decoded_len && decoded[c] != '\r' && decoded[c] != '|'; c++);
-	if (c >= decoded_len)
-	    break;
-
-	if (depth < dirs->cur)
-	    ptrv_remove_range(dirs, depth, dirs->cur);
-
-	oldnode = dirs->buf[dirs->cur-1];
-	if (decoded[c] == '|') {
-	    char *sizestr;
-	    uint64_t size;
-
-	    decoded[c] = '\0';
-    	    sizestr = decoded+c+1;
-    	    for (c++; c < decoded_len && decoded[c] != '\r'; c++);
-	    if (c >= decoded_len)
-	    	break; /* Premature end */
-	    decoded[c] = '\0';
-    	    c++; /* skip LF */
-	    if (!parse_uint64(sizestr, &size))
-	    	continue; /* Skipping bad line */
-
-	    node = new_file_node(name, DC_TYPE_REG, oldnode);
-	    node->u.reg.size = size;
-	} else {
-	    decoded[c] = '\0';
-    	    c++; /* skip LF */
-	    node = new_file_node(name, DC_TYPE_DIR, oldnode);
-	}
-
-	if (node->type == DC_TYPE_REG) {
-	    DCFileList *t;
-	    for (t = oldnode; t != NULL; t = t->parent)
-	    	t->u.dir.totalsize += node->u.reg.size;
-	}
-	if (node->type == DC_TYPE_DIR)
-	    ptrv_append(dirs, node);
-    }
-
-    node = dirs->buf[0];
-    ptrv_free(dirs); /* ignore non-empty */
-    return node;
-}
-
 DCFileList *
 filelist_lookup(DCFileList *node, const char *filename)
 {
@@ -306,9 +234,9 @@ get_sorted_file_list(DCFileList *node, uint32_t *out_count)
     uint32_t c;
 
     assert(node->type == DC_TYPE_DIR);
-    count = hmap_size(node->u.dir.children);
+    count = hmap_size(node->dir.children);
     items = xmalloc((count+1) * sizeof(DCFileList *));
-    hmap_iterator(node->u.dir.children, &it);
+    hmap_iterator(node->dir.children, &it);
     for (c = 0; c < count; c++)
         items[c] = it.next(&it);
     items[count] = NULL;
@@ -335,7 +263,7 @@ filelist_list_recursively(DCFileList *node, char *basepath)
         }
 	free(items);
     } else {
-	screen_putf("%7" PRIu64 "M %s\n", (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(basepath)); /* " */
+	screen_putf("%7" PRIu64 "M %s\n", (uint64_t) (node->size/(1024*1024)), quotearg(basepath)); /* " */
     }
 }
 
@@ -350,23 +278,23 @@ filelist_list(DCFileList *node, bool long_mode)
 
 	maxlen = 0;
 	maxsize = 0;
-	hmap_iterator(node->u.dir.children, &it);
+	hmap_iterator(node->dir.children, &it);
 	while (it.has_next(&it)) {
     	    DCFileList *subnode = it.next(&it);
 
 	    switch (subnode->type) {
 	    case DC_TYPE_REG:
-		maxsize = max(maxsize, subnode->u.reg.size);
+		maxsize = max(maxsize, subnode->size);
         	maxlen = max(maxlen, strlen(subnode->name));
         	break;
             case DC_TYPE_DIR:
-		maxsize = max(maxsize, subnode->u.dir.totalsize);
+		maxsize = max(maxsize, subnode->size);
         	maxlen = max(maxlen, strlen(subnode->name)+1);
         	break;
             }
 	}
     } else {
-    	maxsize = node->u.reg.size;
+    	maxsize = node->size;
     	maxlen = strlen(node->name);
     }
 
@@ -382,16 +310,16 @@ filelist_list(DCFileList *node, bool long_mode)
 	    for (c = 0; items[c] != NULL; c++) {
 		switch (items[c]->type) {
 		case DC_TYPE_REG:
-    		    screen_putf(format, (uint64_t) (items[c]->u.reg.size/(1024*1024)), quotearg(items[c]->name), "");
+    		    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "");
 		    break;
 		case DC_TYPE_DIR:
-    		    screen_putf(format, (uint64_t) (items[c]->u.dir.totalsize/(1024*1024)), quotearg(items[c]->name), "/");
+    		    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "/");
 		    break;
 		}
 	    }
 	    free(items);
 	} else {
-	    screen_putf(format, (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(node->name), "");
+	    screen_putf(format, (uint64_t) (node->size/(1024*1024)), quotearg(node->name), "");
 	}
 	free(format);
     } else {
@@ -440,55 +368,6 @@ filelist_list(DCFileList *node, bool long_mode)
     }
 }
 
-DCFileList *
-filelist_open(const char *filename)
-{
-    struct stat st;
-    uint8_t *contents;
-    char *decoded;
-    uint32_t decoded_len;
-    int fd;
-    DCFileList *root;
-    ssize_t res;
-
-    if (stat(filename, &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
-	return NULL;
-    }
-    contents = malloc(st.st_size);
-    if (contents == NULL) {
-    	screen_putf("%s: %s\n", quotearg(filename), errstr);
-	return NULL;
-    }
-    fd = open(filename, O_RDONLY);
-    if (fd < 0) {
-    	screen_putf(_("%s: Cannot open file for reading - %s\n"), quotearg(filename), errstr);
-	return NULL;
-    }
-    res = full_read(fd, contents, st.st_size);
-    if (res < st.st_size) {
-    	if (res < 0)
-    	    screen_putf(_("%s: Cannot read from file - %s\n"), quotearg(filename), errstr);
-	else
-	    screen_putf(_("%s: Premature end of file\n"), quotearg(filename));	/* XXX: really: file was truncated? */
-	free(contents);
-	if (close(fd) < 0)
-    	    screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
-	return NULL;
-    }
-    if (close(fd) < 0)
-    	screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
-    decoded = huffman_decode(contents, st.st_size, &decoded_len);
-    free(contents);
-    if (decoded == NULL) {
-    	screen_putf(_("%s: Invalid data, cannot decode\n"), quotearg(filename));
-	return NULL;
-    }
-    root = parse_decoded_dclst(decoded, decoded_len);
-    free(decoded);
-    return root;
-}
-
 static void
 dir_to_filelist(DCFileList *parent, const char *path)
 {
@@ -500,8 +379,7 @@ dir_to_filelist(DCFileList *parent, const char *path)
     	screen_putf(_("%s: Cannot open directory - %s\n"), quotearg(path), errstr);
 	return;
     }
-    errno = 0;
-    while ((ep = readdir(dp)) != NULL) {
+    while ((ep = xreaddir(dp)) != NULL) {
     	struct stat st;
 	char *fullname;
 
@@ -519,12 +397,12 @@ dir_to_filelist(DCFileList *parent, const char *path)
 	if (S_ISDIR(st.st_mode)) {
 	    DCFileList *node = new_file_node(ep->d_name, DC_TYPE_DIR, parent);
 	    dir_to_filelist(node, fullname);
-	    parent->u.dir.totalsize += node->u.dir.totalsize;
+	    parent->size += node->size;
 	}
 	else if (S_ISREG(st.st_mode)) {
 	    DCFileList *node = new_file_node(ep->d_name, DC_TYPE_REG, parent);
-	    node->u.reg.size = st.st_size;
-	    parent->u.dir.totalsize += node->u.reg.size;
+	    node->size = st.st_size;
+	    parent->size += node->size;
 	}
 	else {
 	    screen_putf(_("%s: Not a regular file or directory, ignoring\n"), quotearg(fullname));
@@ -544,13 +422,13 @@ filelist_to_string(DCFileList *node, StrBuf *sb, int level)
     	strbuf_append_char_n(sb, level-1, '\t');
 
     if (node->type == DC_TYPE_REG) {
-   	strbuf_appendf(sb, "%s|%" PRIu64 "\r\n", node->name, node->u.reg.size); /* " joe sh bug */
+   	strbuf_appendf(sb, "%s|%" PRIu64 "\r\n", node->name, node->size); /* " joe sh bug */
     } else {
     	HMapIterator it;
 
     	if (level != 0)
    	    strbuf_appendf(sb, "%s\r\n", node->name);
-	hmap_iterator(node->u.dir.children, &it);
+	hmap_iterator(node->dir.children, &it);
 	while (it.has_next(&it)) {
 	    DCFileList *subnode = it.next(&it);
             filelist_to_string(subnode, sb, level+1);
@@ -624,7 +502,7 @@ filelist_create(const char *basedir)
     if (our_filelist != NULL)
     	filelist_free(our_filelist);
     our_filelist = root;
-    my_share_size = our_filelist->u.dir.totalsize;
+    my_share_size = our_filelist->size;
 
     return true;
 }
@@ -988,7 +866,7 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
 static void
 filelist_iterator(DCFileList *node, DCFileListIterator *it)
 {
-    hmap_iterator(node->u.dir.children, &it->it);
+    hmap_iterator(node->dir.children, &it->it);
     it->node = node;
     it->c = 0;
 }
