@@ -25,6 +25,7 @@
 #include <arpa/inet.h>		/* ? */
 #include <unistd.h>		/* POSIX */
 #include <inttypes.h>		/* POSIX */
+#include "iconvme.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
@@ -38,22 +39,37 @@
 #include "common/strleftcmp.h"
 #include "microdc.h"
 
+#define DEFAULT_HUB_RECVQ_SIZE 128
+#define DEFAULT_HUB_SENDQ_SIZE 128
+
+typedef enum {
+    HUB_EXT_NOGETINFO	= 1 << 0,
+    HUB_EXT_NOHELLO	= 1 << 1, 
+} DCHubExtension;
+
 ByteQ *hub_recvq = NULL;
 ByteQ *hub_sendq = NULL;
 int hub_socket = -1;
 DCHubState hub_state = DC_HUB_DISCONNECTED;
 HMap *hub_users = NULL;  /* all users on the current hub (UserInfo->nick => UserInfo) */
+DCLookup *hub_lookup = NULL;
+char *hub_charset = NULL;
+iconv_t to_hub_iconv = (iconv_t) -1;
+iconv_t from_hub_iconv = (iconv_t) -1;
 
 struct sockaddr_in hub_addr;
 static uint32_t hub_recvq_last = 0;
 char *hub_name = NULL;
+static DCHubExtension hub_extensions = 0;
 
 bool
 send_my_info(void)
 {
-    return hub_putf("$MyINFO $ALL %s %s<%s V:%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRIu64 "$|", /* " */
+    /* XXX: hm, H:1/0/0 should be Normal/Registered/Op, calculate this value. */
+    return hub_putf("$MyINFO $ALL %s %s<%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRIu64 "$|", /* " */
 	  my_nick,
-	  my_description, "++", "0.401", is_active ? 'A':'P', my_ul_slots,
+	  my_description,
+	  my_tag, is_active ? 'A':'P', my_ul_slots,
 	  my_speed,
 	  1, /* level, '1' means normal, see DCTC Documentation/Documentation/VAR */
 	  my_email,
@@ -147,21 +163,21 @@ user_info_new(const char *nick)
     info->conn_count = 0;
 
     /* XXX Find existing connections to this user... */
-    ucname = xasprintf(_("%s|UL"), nick);
+    ucname = xasprintf("%s|%s", nick, _("UL"));
     info->conn[info->conn_count] = hmap_get(user_conns, ucname);
     if (info->conn[info->conn_count] != NULL)
     	info->conn_count++;
     free(ucname);
-    ucname = xasprintf(_("%s|DL"), nick);
+    ucname = xasprintf("%s|%s", nick, _("DL"));
     info->conn[info->conn_count] = hmap_get(user_conns, ucname);
     if (info->conn[info->conn_count] != NULL)
     	info->conn_count++;
     free(ucname);
-    /*ucname = xasprintf(_("%s|"), nick);
+    ucname = xasprintf("%s|", nick);
     info->conn[info->conn_count] = hmap_get(user_conns, ucname);
     if (info->conn[info->conn_count] != NULL)
         info->conn_count++;
-    free(ucname);*/
+    free(ucname);
 
     return info;
 }
@@ -224,6 +240,42 @@ hub_putf(const char *format, ...)
     return true;
 }
 
+static void
+hub_address_looked_up(int rc, struct addrinfo *ai, void *data)
+{
+    char *hostname = data;
+
+    hub_lookup = NULL;
+    if (rc != 0) {
+        screen_putf(_("%s: Cannot look up address - %s\n"), quotearg(hostname), gai_strerror(rc));
+        free(data);
+        return;
+    }
+
+    hub_connect((struct sockaddr_in *) ai->ai_addr);
+    free(data);
+}
+
+/* port must be valid */
+void
+hub_new(const char *hostname, uint16_t port)
+{
+    struct sockaddr_in addr;
+
+    if (inet_aton(hostname, &addr.sin_addr)) {
+        addr.sin_family = AF_INET;
+        addr.sin_port = htons(port);
+        hub_connect(&addr); /* Ignore errors */
+    } else {
+        char portstr[6];
+
+        sprintf(portstr, "%" PRIu16, port);
+        screen_putf(_("Looking up IP address for %s\n"), quotearg(hostname));
+        hub_lookup = add_lookup_request(hostname, portstr, NULL, hub_address_looked_up, xstrdup(hostname));
+        hub_state = DC_HUB_LOOKUP;
+    }
+}
+
 void
 hub_connect(struct sockaddr_in *addr)
 {
@@ -236,7 +288,7 @@ hub_connect(struct sockaddr_in *addr)
 
     /* Set non-blocking I/O on socket. */
     if (!fd_set_nonblock_flag(hub_socket, true)) {
-        warn(_("Cannot set socket flags - %s\n"), errstr);
+        warn(_("Cannot set non-blocking flag - %s\n"), errstr);
 	hub_disconnect();
 	return;
     }
@@ -244,7 +296,7 @@ hub_connect(struct sockaddr_in *addr)
     screen_putf(_("Connecting to hub on %s.\n"), sockaddr_in_str(addr));
     if (connect(hub_socket, (struct sockaddr *) addr, sizeof(struct sockaddr_in)) < 0
     	    && errno != EINPROGRESS) {
-        warn(_("Cannot connect - %s\n"), errstr);
+        warn(_("%s: Cannot connect - %s\n"), sockaddr_in_str(addr), errstr);
 	hub_disconnect();
         return;
     }
@@ -257,9 +309,13 @@ hub_connect(struct sockaddr_in *addr)
 void
 hub_disconnect(void)
 {
+    if (hub_state > DC_HUB_DISCONNECTED)
+        screen_putf(_("Shutting down hub connection.\n"));
+    if (hub_lookup != NULL) {
+        cancel_lookup_request(hub_lookup);
+        hub_lookup = NULL;
+    }
     if (hub_socket >= 0) {
-    	if (hub_state > DC_HUB_CONNECT)
-    	    screen_putf(_("Shutting down hub connection.\n"));
     	FD_CLR(hub_socket, &read_fds);
         FD_CLR(hub_socket, &write_fds);
     	if (close(hub_socket) < 0)
@@ -274,10 +330,14 @@ hub_disconnect(void)
     	byteq_clear(hub_sendq);
     if (hub_recvq != NULL)
     	byteq_clear(hub_recvq);
+    hub_recvq_last = 0;
     if (pending_userinfo != NULL) {
     	hmap_foreach_value(pending_userinfo, user_info_free);
     	hmap_clear(pending_userinfo);
     }
+    free(hub_name);
+    hub_name = NULL;
+    hub_extensions = 0;
     hub_state = DC_HUB_DISCONNECTED;
 }
 
@@ -292,6 +352,29 @@ check_state(char *buf, DCUserState state)
     return true;
 }
 
+static void
+parse_hub_extension(const char *ext)
+{
+    if (strcmp(ext, "NoGetINFO") == 0) {
+        hub_extensions |= HUB_EXT_NOGETINFO;
+    } else if (strcmp(ext, "NoHello") == 0) {
+        hub_extensions |= HUB_EXT_NOHELLO;
+    }
+}
+
+static char *
+prepare_chat_string_for_display(const char *str)
+{
+    if (from_hub_iconv != (iconv_t) -1) {
+        char *t1, *t2;
+        t1 = unescape_message(str);
+        t2 = iconv_alloc(from_hub_iconv, t1);
+        free(t1);
+        return t2;
+    }
+    return unescape_message(str);
+}
+
 static void
 hub_handle_command(char *buf, uint32_t len)
 {
@@ -307,6 +390,12 @@ hub_handle_command(char *buf, uint32_t len)
 	    key = buf+len;
 	}
         key = decode_lock(buf+6, key-buf-6, DC_CLIENT_BASE_KEY);
+	if (strleftcmp("EXTENDEDPROTOCOL", buf+6) == 0) {
+	    if (!hub_putf("$Supports NoGetINFO NoHello|")) {
+	        free(key);
+	        return;
+            }
+	}
         if (!hub_putf("$Key %s|", key)) {
 	    free(key);
 	    return;
@@ -316,6 +405,17 @@ hub_handle_command(char *buf, uint32_t len)
 	    return;
 	hub_state = DC_HUB_HELLO;
     }
+    else if (len >= 10 && strncmp(buf, "$Supports ", 10) == 0) {
+        char *p0, *p1;
+
+        hub_extensions = 0;
+        for (p0 = buf+10; (p1 = strchr(p0, ' ')) != NULL; p0 = p1+1) {
+            *p1 = '\0';
+            parse_hub_extension(p0);
+        }
+        if (*p0 != '\0')
+            parse_hub_extension(p0);
+    }
     else if (strcmp(buf, "$GetPass") == 0) {
     	if (my_password == NULL) {
     	    screen_putf(_("Hub requires password.\n"));
@@ -372,7 +472,6 @@ hub_handle_command(char *buf, uint32_t len)
 	    	return;
 	    if (!hub_putf("$GetNickList|"))
 	    	return;
-	    /* XXX: hm, /0/0 should be Registered/Op, calculate this value. */
 	    if (!send_my_info())
 	    	return;
     	    hub_state = DC_HUB_LOGGED_IN;
@@ -380,8 +479,11 @@ hub_handle_command(char *buf, uint32_t len)
 	    flag_putf(DC_DF_JOIN_PART, _("User %s logged in.\n"), quotearg(buf+7));
 	    ui = user_info_new(buf+7);
             hmap_put(hub_users, ui->nick, ui);
-	    if (!hub_putf("$GetINFO %s|", buf+7))
-	    	return;
+	    if ((hub_extensions & HUB_EXT_NOGETINFO) == 0) {
+	        if (!hub_putf("$GetINFO %s %s|", buf+7, my_nick))
+	    	    return;
+                ui->info_quered = true;
+            }
 	}
     }
     else if (len >= 8 && strncmp(buf, "$MyINFO ", 8) == 0) {
@@ -468,7 +570,9 @@ hub_handle_command(char *buf, uint32_t len)
 	/* hub_disconnect(); */
     }
     else if (len >= 3 && (buf[0] == '<' || strncmp(buf, " * ", 3) == 0)) {
-	char *end;
+	char *head;
+	char *tail;
+	/*char *msg;*/
 	bool first = true;
 	/*int scrwidth;
 	size_t firstlen;
@@ -477,27 +581,31 @@ hub_handle_command(char *buf, uint32_t len)
         screen_get_size(NULL, &scrwidth);
         firstlen = scrwidth - strlen(_("Public:"));
         otherlen = scrwidth - strlen(_(" | "));*/
-	for (; (end = strchr(buf, '\n')) != NULL; buf = end+1) {
+        buf = prepare_chat_string_for_display(buf);
+	for (head = buf; (tail = strchr(head, '\n')) != NULL; head = tail+1) {
 	    /*PtrV *wrapped;*/
 
-	    if (end[-1] == '\r') /* end > buf here, buf[0] == '<' or ' ' */
-		end[-1] = '\0';
+	    if (tail[-1] == '\r') /* end > buf here, buf[0] == '<' or ' ' */
+		tail[-1] = '\0';
 	    else
-		end[0] = '\0';
+		tail[0] = '\0';
 
             /*wrapped = wordwrap(quotearg(buf), first ? firstlen : otherlen, otherlen);
             for (c = 0; c < wrapped->cur; c++)
 	        flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), );
             ptrv_foreach(wrapped, free);
             ptrv_free(wrapped);*/
-	    flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(buf));
+	    flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(head));
             first = false;
 	}
-	flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(buf));
+	flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(head));
+	free(buf);
     }
     else if (len >= 5 && strncmp(buf, "$To: ", 5) == 0) {
     	char *msg;
-	char *end;
+	char *tail;
+	char *frm;
+	char *head;
 	bool first = true;
 
 	msg = strchr(buf+5, '$');
@@ -508,23 +616,30 @@ hub_handle_command(char *buf, uint32_t len)
         *msg = '\0';
 	msg++;
 
-	for (; (end = strchr(msg, '\n')) != NULL; msg = end+1) {
-	    if (end[-1] == '\r') /* end > buf here, msg[0] == '<' or ' ' */
-		end[-1] = '\0';
+	/* FIXME: WTF is this? Remove multiple "From: "? Why!? */
+	frm = buf+5;
+	while ((tail = strstr(msg, "From: ")) != NULL && tail < msg)
+	    frm = tail+6;
+
+        msg = prepare_chat_string_for_display(msg);
+	for (head = msg; (tail = strchr(head, '\n')) != NULL; head = tail+1) {
+	    if (tail[-1] == '\r') /* tail > buf here because head[0] == '<' or ' ' */
+		tail[-1] = '\0';
 	    else
-		end[0] = '\0';
+		tail[0] = '\0';
 	    if (first) {
-		screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, buf+5), quotearg_n(1, msg));
+		screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, frm), quotearg_n(1, head));
 		first = false;
 	    } else {
-		screen_putf(_(" | %s\n"), quotearg(msg));
+		screen_putf(_(" | %s\n"), quotearg(head));
 	    }
 	}
 	if (first) {
-	    screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, buf+5), quotearg_n(1, msg));
+	    screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, frm), quotearg_n(1, head));
 	} else {
-	    screen_putf(_(" | %s\n"), quotearg(msg));
+	    screen_putf(_(" | %s\n"), quotearg(head));
 	}
+	free(msg);
     }
     else if (len >= 13 && strncmp(buf, "$ConnectToMe ", 13) == 0) {
 	struct sockaddr_in addr;
@@ -607,7 +722,7 @@ hub_handle_command(char *buf, uint32_t len)
 	    	ui = user_info_new(nick);
                 hmap_put(hub_users, ui->nick, ui);
 	    }
-	    if (!ui->info_quered) {
+	    if (!ui->info_quered && (hub_extensions & HUB_EXT_NOGETINFO) == 0) {
 		if (!hub_putf("$GetINFO %s %s|", nick, my_nick))
 		    return;
 		ui->info_quered = true;
@@ -627,7 +742,7 @@ hub_handle_command(char *buf, uint32_t len)
                 ui = user_info_new(nick);
                 hmap_put(hub_users, ui->nick, ui);
 	    }
-	    if (!ui->info_quered) {
+	    if (!ui->info_quered && (hub_extensions & HUB_EXT_NOGETINFO) == 0) {
 	    	if (!hub_putf("$GetINFO %s %s|", nick, my_nick))
 		    return;
 		ui->info_quered = true;
@@ -744,6 +859,7 @@ hub_now_writable(void)
         }
         if (error != 0) { /* The connect call on the socket failed. */
 	    warn(_("Cannot connect - %s\n"), strerror(error) /* not errno! */);
+	    /* XXX: need to know hub address to put in error message */
 	    hub_disconnect();
 	    return;
 	}
