@@ -43,6 +43,7 @@
 #include "xstrndup.h"		/* Gnulib */
 #include "dirname.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
+#include "human.h"		/* Gnulib */
 #include "getline.h"		/* Gnulib/GNU Libc */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
@@ -51,7 +52,7 @@
 #include "common/strleftcmp.h"
 #include "common/tempdir.h"
 #include "common/tempfailure.h"
-#include "ipc.h"
+#include "common/msgq.h"
 #include "microdc.h"
 
 enum {
@@ -97,12 +98,13 @@ static int signal_pipe[2] = { -1, -1 };
 static int listen_socket = -1;
 static int search_socket = -1;
 
-static const char *short_opts = "n";
+static const char *short_opts = "c:n";
 static struct option long_opts[] = {
+    { "config", required_argument, NULL, 'c' },
     { "no-config", no_argument, NULL, 'n' },
     { "version", no_argument, NULL, VERSION_OPT },
     { "help", no_argument, NULL, HELP_OPT },
-    { 0, 0, 0, 0 }
+    { 0, }
 };
 
 void
@@ -120,17 +122,24 @@ warn_file_error(int res, bool write, const char *filename)
 }
 
 void
-warn_socket_error(int res, bool write, const char *subject)
+warn_socket_error(int res, bool write, const char *format, ...)
 {
+    char *subject;
+    va_list args;
+    
+    va_start(args, format);
+    subject = xvasprintf(format, args);
+    va_end(args);
     if (write) {
-	warn("Cannot send to %s - %s\n", quotearg(subject), errstr);
+	warn("Cannot send to %s - %s\n", subject, errstr);
     } else {
 	if (res < 0) {
-	    warn("Cannot receive from %s - %s\n", quotearg(subject), errstr);
+	    warn("Cannot receive from %s - %s\n", subject, errstr);
 	} else {
-	    warn("Disconnected from %s.\n", quotearg(subject));
+	    warn("Disconnected from %s.\n", subject);
 	}
     }
+    free(subject);
 }
 
 bool
@@ -140,46 +149,6 @@ get_package_file(const char *name, char **outname)
     return true;
 }
 
-static bool
-validate_nick(DCUserConn *uc, const char *nick)
-{
-    DCUserInfo *ui;
-
-    ui = hmap_get(pending_userinfo, nick);
-    if (ui != NULL) {
-    	if (ui->conn_count < DC_USER_MAX_CONN) {
-	    uc->info = ui;
-	    uc->info->refcount++;
-	    uc->info->conn[uc->info->conn_count++] = uc;
-	    hmap_remove(pending_userinfo, ui->nick);
-	    user_info_free(ui);
-	    /*hmap_remove(user_conns, uc->name);
-	    ptrv_append(user_conn_unknown_free, uc->name);
-	    new_name = xasprintf("%s|", uc->info->nick);
-	    uc->name = new_name;
-	    hmap_put(user_conns, uc->name, uc);*/
-	    return true;
-	}
-        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
-        hmap_remove(pending_userinfo, ui->nick);
-	user_info_free(ui);
-	return false;
-    }
-
-    ui = hmap_get(hub_users, nick);
-    if (ui != NULL) {
-    	if (ui->conn_count < DC_USER_MAX_CONN) {
-    	    uc->info = ui;
-	    uc->info->refcount++;
-	    uc->info->conn[uc->info->conn_count++] = uc;
-	    return true;
-	}
-        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
-    }
-
-    return false;
-}
-
 bool
 has_user_conn(DCUserInfo *ui, DCTransferDirection dir)
 {
@@ -191,25 +160,74 @@ has_user_conn(DCUserInfo *ui, DCTransferDirection dir)
     return false;
 }
 
+static void
+update_user_connection_name(DCUserConn *uc, const char *format, ...)
+{
+    char *newname;
+    va_list args;
+    
+    va_start(args, format);
+    newname = xvasprintf(format, args);
+    va_end(args);
+
+    hmap_remove(user_conns, uc->name);
+    flag_putf(DC_DF_CONNECTIONS, _("User connection %s renamed to %s.\n"), quote_n(0, uc->name), quote_n(1, newname));
+    if (strchr(uc->name, '|') == NULL)
+        ptrv_append(user_conn_unknown_free, uc->name);
+    else
+        free(uc->name);
+    uc->name = newname;
+    hmap_put(user_conns, uc->name, uc);
+}
+
+static bool
+validate_nick(DCUserConn *uc, const char *nick)
+{
+    DCUserInfo *ui;
+
+    ui = hmap_remove(pending_userinfo, nick);
+    if (ui != NULL) {
+    	if (ui->conn_count < DC_USER_MAX_CONN) {
+	    uc->info = ui;
+	    uc->info->refcount++;
+	    uc->info->conn[uc->info->conn_count++] = uc;
+	    hmap_remove(pending_userinfo, ui->nick);
+	    user_info_free(ui);
+	    update_user_connection_name(uc, "%s|", uc->info->nick);
+	    return true;
+	}
+        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
+	user_info_free(ui);
+	return false;
+    }
+
+    ui = hmap_get(hub_users, nick);
+    if (ui != NULL) {
+    	if (ui->conn_count < DC_USER_MAX_CONN) {
+    	    uc->info = ui;
+	    uc->info->refcount++;
+	    uc->info->conn[uc->info->conn_count++] = uc;
+	    update_user_connection_name(uc, "%s|", uc->info->nick);
+	    return true;
+	}
+        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
+        return false;
+    }
+
+    return false;
+}
+
 static bool
 validate_direction(DCUserConn *uc, DCTransferDirection dir)
 {
-    char *newname;
-
     if (uc->dir != DC_DIR_UNKNOWN)
     	return false; /* Shouldn't happen, but better be sure */
 
     if (has_user_conn(uc->info, dir))
     	return false;
 
-    hmap_remove(user_conns, uc->name);
-    ptrv_append(user_conn_unknown_free, uc->name);
-    newname = xasprintf(
-	    dir == DC_DIR_SEND ? _("%s|UL") : _("%s|DL"), uc->info->nick);
-    flag_putf(DC_DF_CONNECTIONS, _("User %s validated as %s\n"), quotearg_n(0, uc->name), quotearg_n(1, newname));
-    uc->name = newname;
     uc->dir = dir;
-    hmap_put(user_conns, uc->name, uc);
+    update_user_connection_name(uc, "%s|%s", uc->info->nick, dir == DC_DIR_SEND ? _("UL") : _("DL"));
     /* Don't check for a free slot here. We will do that later. */
 
     /* Make new connection if we want to download but couldn't. */
@@ -223,41 +241,64 @@ validate_direction(DCUserConn *uc, DCTransferDirection dir)
 }
 
 static void
-handle_user_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
+display_transfer_ended_msg(bool upload, DCUserConn *uc, bool success, const char *extras_fmt, ...)
 {
-    DCUserConn *uc = userdata;
+    char timebuf[LONGEST_ELAPSED_TIME+1];
+    char ratebuf[LONGEST_HUMAN_READABLE+1];
+    char sizebuf[LONGEST_HUMAN_READABLE+1];
+    char *str1;
+    char *str2;
+    time_t now;
+    uint64_t len;
+    va_list args;
 
-    switch (error) {
-    case IPC_ERROR_WRITE:
-    	warn_socket_error(result, true, _("user process"));
-	break;
-    case IPC_ERROR_READ:
-    	warn_socket_error(result, false, _("user process"));
-	break;
-    case IPC_ERROR_DATA:
-    	warn(_("Logical communication error with user process (during message %d).\n"), result);
-    	break;
+    len = uc->transfer_pos - uc->transfer_start;
+
+    if (uc->transfer_time == (time_t) -1) {
+        str2 = xstrdup("");
+    } else {
+	time(&now);
+	if (now == (time_t) -1) {
+	    warn(_("Cannot get current time - %s\n"), errstr);
+            str2 = xstrdup("");
+	} else {
+	    str2 = xasprintf(_(" in %s (%s/s)"),
+	                elapsed_time_to_string(now - uc->transfer_time, timebuf),
+	                human_readable(
+	                    len/MAX(now - uc->transfer_time, 1), ratebuf,
+                            human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B,
+                            1, 1));
+	}
     }
-    hmap_remove(user_conns, uc->name);
-    user_disconnect(uc);
+
+    va_start(args, extras_fmt);
+    str1 = xvasprintf(extras_fmt, args);
+    va_end(args);
+
+    flag_putf(upload ? DC_DF_UPLOAD : DC_DF_DOWNLOAD,
+	_("%s: %s of %s %s%s. %s %s%s.\n"),
+	quotearg_n(0, uc->info->nick),
+	upload ? _("Upload") : _("Download"),
+	quote_n(1, base_name(uc->transfer_file)),
+	success ? _("succeeded") : _("failed"),
+	str1,
+        human_readable(len, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1),
+	ngettext("transferred", "transferred", len),
+	str2);
+
+    free(str1);
+    free(str2);
 }
 
 static void
-handle_ended_upload(DCUserConn *uc, bool success)
+handle_ended_upload(DCUserConn *uc, bool success, const char *reason)
 {
-    uint64_t len = uc->transfer_pos - uc->transfer_start;
-
-    bytes_sent += len;
+    bytes_sent += uc->transfer_pos - uc->transfer_start;
     if (uc->occupied_slot) {
         /* Only display "Upload failed" if "Starting upload" was
          * displayed previously.
          */
-        flag_putf(DC_DF_UPLOAD, _("%s: Upload of %s %s. %" PRIu64 " %s.\n"),
-            quotearg_n(0, uc->info->nick),
-            quote_n(1, base_name(uc->transfer_file)),
-            success ? _("succeeded") : _("failed"),
-            len,
-            ngettext("byte transferred", "bytes transferred", len));
+	display_transfer_ended_msg(true, uc, success, " (%s)", reason);
         used_ul_slots--;
         uc->occupied_slot = false;
     }
@@ -271,9 +312,7 @@ handle_ended_upload(DCUserConn *uc, bool success)
 static void
 handle_ended_download(DCUserConn *uc, bool success, const char *reason)
 {
-    uint64_t len = uc->transfer_pos - uc->transfer_start;
-
-    bytes_received += len;
+    bytes_received += uc->transfer_pos - uc->transfer_start;
     if (uc->occupied_slot) {
         used_dl_slots--;
         uc->occupied_slot = false;
@@ -292,20 +331,16 @@ handle_ended_download(DCUserConn *uc, bool success, const char *reason)
                 if (browse_user != NULL && strcmp(browse_user->nick, uc->info->nick) == 0 && browse_list == NULL) {
                     char *filename;
 
-                    screen_putf(_("Now browsing %s.\n"), quotearg(browse_user->nick));
                     filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, browse_user->nick);
-                    browse_list = filelist_open(filename);
-                    browse_path = xstrdup("/");
-                    browse_path_previous = NULL;
-                    free(filename);
-                    update_prompt();
+		    add_parse_request(browse_list_parsed, filename, xstrdup(browse_user->nick));
+		    free(filename);
                 }
             } else {
                 char *local_file = resolve_download_file(uc->info, queued);
                 char *final_file = xstrndup(local_file, strlen(local_file)-5);
                 if (safe_rename(local_file, final_file) != 0) {
                     warn(_("%s: Cannot rename file to %s - %s\n"), quotearg_n(0, local_file), quote_n(1, final_file), errstr);
-                    reason = "cannot rename file"; /* XXX: would like a more elaborate error message here perhaps? */
+                    reason = _("cannot rename file"); /* XXX: would like a more elaborate error message here perhaps? */
                     queued->status = DC_QS_ERROR;
                     success = false;
                 }
@@ -315,19 +350,9 @@ handle_ended_download(DCUserConn *uc, bool success, const char *reason)
         } else {
             queued->status = DC_QS_ERROR;
         }
-        flag_putf(DC_DF_DOWNLOAD, _("%s: Download of %s %s (%s). %" PRIu64 " %s.\n"),
-            quotearg_n(0, uc->info->nick),
-            quote_n(1, base_name(uc->transfer_file)),
-            success ? _("succeeded") : _("failed"),
-            reason, len,
-            ngettext("byte transferred", "bytes transferred", len));
+	display_transfer_ended_msg(false, uc, success, " (%s)", reason);
     } else {
-        flag_putf(DC_DF_DOWNLOAD, _("%s: Download of %s %s but unqueued (%s). %" PRIu64 " %s.\n"),
-            quotearg_n(0, uc->info->nick),
-            quote_n(1, base_name(uc->transfer_file)),
-            success ? _("succeeded") : _("failed"),
-            reason, len,
-            ngettext("byte transferred", "bytes transferred", len));
+	display_transfer_ended_msg(false, uc, success, " but unqueued (%s)", reason);
     }
     free(uc->transfer_file);
     uc->transfer_file = NULL;
@@ -336,201 +361,38 @@ handle_ended_download(DCUserConn *uc, bool success, const char *reason)
     uc->transferring = false;
 }
 
-static bool
-handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
-{
-    DCUserConn *uc = userdata;
-
-    switch (id) {
-    case DC_MSG_SCREEN_PUT: {
-    	char *msg;
-    	uint32_t flag;
-
-    	ipc_get(ipc, id, IPC_INT32, &flag, IPC_STR, &msg, IPC_END);
-    	flag_putf(flag, _("User %s: %s"), quotearg(uc->name), msg); /* msg already quoted */
-	free(msg);
-    	return true;
-    }
-    case DC_MSG_WANT_DOWNLOAD: {
-    	bool reply;
-
-    	ipc_get(ipc, id, IPC_END);
-    	reply = !has_user_conn(uc->info, DC_DIR_RECEIVE)
-    	        && (uc->queue_pos < uc->info->download_queue->cur);
-	return ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
-    }
-    case DC_MSG_VALIDATE_DIR: {
-    	DCTransferDirection dir;
-	bool reply;
-
-    	if (!ipc_get(ipc, id, IPC_INT, &dir, IPC_END))
-	    return false;
-    	reply = validate_direction(uc, dir);
-	return ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
-    }
-    case DC_MSG_VALIDATE_NICK: {
-    	char *nick;
-	bool reply;
-    
-    	if (!ipc_get(ipc, id, IPC_STR, &nick, IPC_END))
-	    return false;
-    	reply = validate_nick(uc, nick);
-	free(nick);
-	if (uc->info != NULL) {
-	    /* The user can only be DC_ACTIVE_SENT_PASSIVE if we are passive.
-	     * So if we managed to connect to them even though we were passive,
-	     * they must be active.
-	     */
-	    if (uc->info->active_state == DC_ACTIVE_SENT_PASSIVE)
-		uc->info->active_state = DC_ACTIVE_KNOWN_ACTIVE;
-	    /* The user can only be DC_ACTIVE_SENT_ACTIVE if we are active.
-	     * We must set to DC_ACTIVE_UNKNOWN because otherwise
-	     * hub.c:hub_connect_user might say "ConnectToMe already sent to
-	     * user" next time we're trying to connect to them.
-	     */
-	    if (uc->info->active_state == DC_ACTIVE_SENT_ACTIVE)
-	    	uc->info->active_state = DC_ACTIVE_UNKNOWN;
-	}
-        return ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
-    }
-    case DC_MSG_GET_MY_NICK: {
-        return ipc_get(ipc, id, IPC_END) && ipc_put(ipc, id, IPC_STR, my_nick, IPC_END);
-    }
-    case DC_MSG_TRANSFER_STATUS: {
-    	return ipc_get(uc->ipc, id, IPC_INT64, &uc->transfer_pos, IPC_END);
-    }
-    case DC_MSG_TRANSFER_START: {
-    	if (!ipc_get(ipc, id, IPC_INT64, &uc->transfer_start, IPC_INT64, &uc->transfer_total, IPC_END))
-	    return false;
-        if (uc->transfer_start != 0) {
-	    flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " of %" PRIu64 " %s).\n"),
-	            quotearg_n(0, uc->info->nick),
-                    uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
-                    quote_n(1, base_name(uc->transfer_file)),
-                    uc->transfer_total - uc->transfer_start,
-                    uc->transfer_total,
-                    ngettext("byte", "bytes", uc->transfer_total));
-        } else {
-	    flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " %s).\n"),
-                    quotearg_n(0, uc->info->nick),
-                    uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
-                    quote_n(1, base_name(uc->transfer_file)),
-                    uc->transfer_total,
-                    ngettext("byte", "bytes", uc->transfer_total));
-        }
-        uc->transferring = true;
-        uc->transfer_pos = uc->transfer_start;
-        uc->transfer_time = time(NULL);
-        if (uc->transfer_time == (time_t) -1)
-            warn(_("Cannot get current time - %s\n"), errstr);
-	return true;
-    }
-    case DC_MSG_CHECK_DOWNLOAD: {
-    	if (!ipc_get(ipc, id, IPC_END))
-    	    return false;
-
-        for (; uc->queue_pos < uc->info->download_queue->cur; uc->queue_pos++) {
-    	    DCQueuedFile *queued;
-    	    char *local_file;
-    	    bool result;
-
-            queued = uc->info->download_queue->buf[uc->queue_pos];
-            if (queued->status != DC_QS_DONE) {
-                local_file = resolve_download_file(uc->info, queued);
-                uc->queued_valid = true;
-                uc->transfer_file = xstrdup(queued->filename);
-                uc->occupied_slot = true;
-                queued->status = DC_QS_PROCESSING;
-                used_dl_slots++;
-                result = ipc_put(ipc, id, IPC_STR, local_file, IPC_STR, uc->transfer_file, IPC_INT64, queued->length, IPC_END);
-                free(local_file);
-		return result;
-            }
-        }
-        return ipc_put(ipc, id, IPC_STR, NULL, IPC_STR, NULL, IPC_INT64, (uint64_t) 0, IPC_END);
-    }
-    case DC_MSG_DOWNLOAD_ENDED: {
-        bool success;
-        char *reason;
-
-    	if (!ipc_get(ipc, id, IPC_BOOL, &success, IPC_STR, &reason, IPC_END))
-	    return false;
-        handle_ended_download(uc, success, reason);
-        free(reason);
-    	return true;
-    }
-    case DC_MSG_CHECK_UPLOAD: {
-        char *remote_file;
-        char *local_file;
-        bool result;
-
-    	if (!ipc_get(ipc, id, IPC_STR, &remote_file, IPC_END))
-	    return false;
-        if (used_ul_slots >= my_ul_slots && !uc->info->slot_granted) {
-            uc->transfer_file = NULL;
-            uc->occupied_slot = false;
-            local_file = NULL;
-            free(remote_file);
-        } else {
-            uc->transfer_file = remote_file;
-            uc->occupied_slot = true;
-            used_ul_slots++;
-            local_file = resolve_upload_file(uc->info, remote_file);
-        }
-        result = ipc_put(ipc, id, IPC_BOOL, uc->occupied_slot, IPC_STR, local_file, IPC_END);
-        free(local_file);
-        return result;
-    };
-    case DC_MSG_UPLOAD_ENDED: {
-        bool success;
-
-        if (!ipc_get(ipc, id, IPC_BOOL, &success, IPC_END))
-            return false;
-        handle_ended_upload(uc, success);
-        return true;
-    };
-    case DC_MSG_DISCONNECT:
-        /* The DISCONNECT message will not be sent from users, only to users. */
-    default:
-        warn(_("Received unknown message %d from user process, shutting down process.\n"), id);
-    	hmap_remove(user_conns, uc->name);
-	user_disconnect(uc);
-	return false;
-    }
-}
-
 DCUserConn *
 user_connection_new(struct sockaddr_in *addr, int user_socket)
 {
     DCUserConn *uc;
-    int main_sockets[2];
+    int get_fd[2] = { -1, -1 };
+    int put_fd[2] = { -1, -1 };
     pid_t pid;
 
-    if (socketpair(AF_LOCAL, SOCK_STREAM, 0, main_sockets) < 0) {
-    	warn(_("Cannot create socket pair - %s\n"), errstr);
-	return NULL;
+    if (pipe(get_fd) != 0 || pipe(put_fd) != 0) {
+        warn(_("Cannot create pipe pair - %s\n"), errstr);
+        goto cleanup;
     }
 
     pid = fork();
     if (pid < 0) {
     	warn(_("Cannot create process - %s\n"), errstr);
-	if (close(main_sockets[0]) < 0)
-	    warn(_("Cannot close socket - %s\n"), errstr);
-	if (close(main_sockets[1]) < 0)
-	    warn(_("Cannot close socket - %s\n"), errstr);
-	return NULL;
+    	goto cleanup;
     }
     if (pid == 0)
-	user_main(main_sockets, addr, user_socket);
+	user_main(put_fd, get_fd, addr, user_socket);
 
-    if (close(main_sockets[1]) < 0)
-    	warn(_("Cannot close socket - %s\n"), errstr);
+    if (close(get_fd[1]) != 0 || close(put_fd[0]) != 0)
+        warn(_("Cannot close pipe - %s\n"), errstr);
+    /* Non-blocking mode is not required, but there may be some latency otherwise. */
+    if (!fd_set_nonblock_flag(get_fd[0], true) || !fd_set_nonblock_flag(put_fd[1], true))
+        warn(_("Cannot set non-blocking flag - %s\n"), errstr);
+    /* The user socket is only used by the newly created user process. */
     if (user_socket >= 0 && close(user_socket) < 0)
     	warn(_("Cannot close socket - %s\n"), errstr);
 
     uc = xmalloc(sizeof(DCUserConn));
     uc->pid = pid;
-    uc->main_socket = main_sockets[0];
     uc->info = NULL;
     uc->occupied_slot = 0;
     uc->dir = DC_DIR_UNKNOWN;
@@ -543,25 +405,50 @@ user_connection_new(struct sockaddr_in *addr, int user_socket)
     uc->queue_pos = 0;
     uc->queued_valid = false;
     /* uc->we_connected = (user_socket < 0); */
-    add_user_conn(uc);
-    FD_SET(uc->main_socket, &read_fds);
-
+    uc->get_mq = msgq_new(get_fd[0]);
+    uc->put_mq = msgq_new(put_fd[1]);
+    FD_SET(uc->get_mq->fd, &read_fds);
+    if (user_conn_unknown_free->cur > 0) {
+        uc->name = ptrv_remove_first(user_conn_unknown_free);
+    } else {
+        /* TRANSLATORS: This represents the connection name used when
+         * the user name is not yet known. It must not contains '|',
+         * because that's used to distinguish between 'unknown' and
+         * (perhaps partially) identified connections.
+         */
+        uc->name = xasprintf(_("unknown%" PRIu32), user_conn_unknown_last+1);
+        user_conn_unknown_last++;
+    }
+    hmap_put(user_conns, uc->name, uc);
     return uc;
+
+  cleanup:
+    if (get_fd[0] != -1)
+        close(get_fd[0]);
+    if (get_fd[1] != -1)
+        close(get_fd[1]);
+    if (put_fd[0] != -1)
+        close(put_fd[0]);
+    if (put_fd[1] != -1)
+        close(put_fd[1]);
+    return NULL;
 }
 
 void
 user_disconnect(DCUserConn *uc)
 {
+    flag_putf(DC_DF_CONNECTIONS, _("Shutting down user connection process for %s.\n"), quote(uc->name)); /* XXX: move where? */
+
+    hmap_remove(user_conns, uc->name);
+
     if (uc->occupied_slot) { /* could also check that uc->transfer_file != NULL */
         if (uc->dir == DC_DIR_SEND) {
-            handle_ended_upload(uc, false);
+            handle_ended_upload(uc, false, "connection terminated prematurely");
         } else { /* uc->dir == DC_DIR_RECV */
             handle_ended_download(uc, false, "connection terminated prematurely");
         }
     }
 
-    flag_putf(DC_DF_CONNECTIONS, _("User %s: Shutting down user connection.\n"), quotearg(uc->name)); /* XXX: move where? */
-    ipc_free(uc->ipc);
     if (uc->info != NULL) {
     	int c;
 	for (c = 0; c < uc->info->conn_count; c++) {
@@ -575,10 +462,19 @@ user_disconnect(DCUserConn *uc)
 	    uc->info->conn[c] = uc->info->conn[c+1];
 	user_info_free(uc->info);
     }
-    FD_CLR(uc->main_socket, &read_fds);
-    FD_CLR(uc->main_socket, &write_fds);
-    if (close(uc->main_socket) < 0)
-	warn(_("Cannot close socket - %s\n"), errstr);
+
+    FD_CLR(uc->get_mq->fd, &read_fds);
+    FD_CLR(uc->put_mq->fd, &write_fds);
+    if (close(uc->get_mq->fd) != 0 || close(uc->put_mq->fd) != 0)
+	warn(_("Cannot close pipe - %s\n"), errstr);
+    msgq_free(uc->get_mq);
+    uc->get_mq = NULL;
+    msgq_free(uc->put_mq);
+    uc->put_mq = NULL;
+    if (strchr(uc->name, '|') == NULL)
+        ptrv_append(user_conn_unknown_free, uc->name);
+    else
+        free(uc->name);
     free(uc);
 }
 
@@ -605,43 +501,210 @@ user_conn_status_to_string(DCUserConn *uc, time_t now)
     return status;
 }
 
-void
-add_user_conn(DCUserConn *uc)
-{
-    uc->ipc = ipc_new(uc->main_socket, handle_user_msg, handle_user_ipc_error, uc);
-    if (user_conn_unknown_free->cur > 0) {
-    	uc->name = ptrv_remove_first(user_conn_unknown_free);
-    } else {
-	uc->name = xasprintf(_("unknown%d"), user_conn_unknown_last+1);
-	user_conn_unknown_last++;
-    }
-    hmap_put(user_conns, uc->name, uc);
-}
-
+/* This function should be called when the user process it to be notified
+ * about the termination (when it isn't terminating by itself).
+ */
 void
 user_conn_cancel(DCUserConn *uc)
 {
-    screen_putf(_("Disconnecting user.\n"));
-
-    if (!ipc_put(uc->ipc, DC_MSG_DISCONNECT, IPC_END))
-    	return;
-    if (!ipc_get(uc->ipc, DC_MSG_DISCONNECT, IPC_END))
-    	return;
-
-    /* We alter the map while iterating over it, but that's OK since we are removing the current value. */
-    hmap_remove(user_conns, uc->name);
-    user_disconnect(uc);
+    user_disconnect(uc); /* MSG: "connection terminated by request" */
 }
 
-/*bool
-get_user_conn_status(DCUserConn *uc)
+static void
+user_request_fd_writable(DCUserConn *uc)
 {
-    if (!ipc_put(uc->ipc, DC_MSG_STATUS, IPC_END))
-    	return false;
-    if (!ipc_get(uc->ipc, DC_MSG_STATUS, IPC_INT64, &uc->transfer_pos, IPC_END))
-    	return false;
-    return true;
-}*/
+    int res;
+
+    res = msgq_write(uc->put_mq);
+    if (res == 0 || (res < 0 && errno != EAGAIN)) {
+        warn_socket_error(res, true, _("user process %s"), quote(uc->name));
+        user_disconnect(uc); /* MSG: socket error above */
+        return;
+    }
+    if (!msgq_has_partial_msg(uc->put_mq))
+        FD_CLR(uc->put_mq->fd, &write_fds);
+}
+
+static void
+user_result_fd_readable(DCUserConn *uc)
+{
+    int res;
+
+    res = msgq_read(uc->get_mq);
+    if (res == 0 || (res < 0 && errno != EAGAIN)) {
+        warn_socket_error(res, false, _("user process %s"), quote(uc->name));
+        user_disconnect(uc); /* MSG: socket error above */
+        return;
+    }
+    while (msgq_has_complete_msg(uc->get_mq)) {
+	int id;
+
+        msgq_peek(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
+	switch (id) {
+	case DC_MSG_SCREEN_PUT: {
+            char *msg;
+            uint32_t flag;
+
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT32, &flag, MSGQ_STR, &msg, MSGQ_END);
+            flag_putf(flag, _("User %s: %s"), quotearg(uc->name), msg); /* msg already quoted */
+            free(msg);
+            break;
+        }
+        case DC_MSG_WANT_DOWNLOAD: {
+    	    bool reply;
+
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
+    	    reply = !has_user_conn(uc->info, DC_DIR_RECEIVE)
+    	            && (uc->queue_pos < uc->info->download_queue->cur);
+            msgq_put(uc->put_mq, MSGQ_BOOL, reply, MSGQ_END);
+            FD_SET(uc->put_mq->fd, &write_fds);
+            break;
+        }
+        case DC_MSG_VALIDATE_DIR: {
+    	    DCTransferDirection dir;
+    	    bool reply;
+
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT, &dir, MSGQ_END);
+    	    reply = validate_direction(uc, dir);
+    	    msgq_put(uc->put_mq, MSGQ_BOOL, reply, MSGQ_END);
+    	    FD_SET(uc->put_mq->fd, &write_fds);
+    	    break;
+        }
+        case DC_MSG_VALIDATE_NICK: {
+    	    char *nick;
+    	    bool reply;
+    
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_STR, &nick, MSGQ_END);
+    	    reply = validate_nick(uc, nick);
+    	    free(nick);
+            if (uc->info != NULL) {
+	        /* The user can only be DC_ACTIVE_SENT_PASSIVE if we are passive.
+	         * So if we managed to connect to them even though we were passive,
+	         * they must be active.
+	         */
+	        if (uc->info->active_state == DC_ACTIVE_SENT_PASSIVE)
+		    uc->info->active_state = DC_ACTIVE_KNOWN_ACTIVE;
+                /* The user can only be DC_ACTIVE_SENT_ACTIVE if we are active.
+                 * We must set to DC_ACTIVE_UNKNOWN because otherwise
+                 * hub.c:hub_connect_user might say "ConnectToMe already sent to
+                 * user" next time we're trying to connect to them.
+                 */
+                 if (uc->info->active_state == DC_ACTIVE_SENT_ACTIVE)
+                     uc->info->active_state = DC_ACTIVE_UNKNOWN;
+            }
+            msgq_put(uc->put_mq, MSGQ_BOOL, reply, MSGQ_END);
+            FD_SET(uc->put_mq->fd, &write_fds);
+            break;
+        }
+        case DC_MSG_GET_MY_NICK:
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
+            msgq_put(uc->put_mq, MSGQ_STR, my_nick, MSGQ_END);
+            FD_SET(uc->put_mq->fd, &write_fds);
+            break;
+        case DC_MSG_TRANSFER_STATUS:
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT64, &uc->transfer_pos, MSGQ_END);
+    	    break;
+        case DC_MSG_TRANSFER_START:
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT64, &uc->transfer_start, MSGQ_INT64, &uc->transfer_total, MSGQ_END);
+    	    if (uc->transfer_start != 0) {
+	        flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " of %" PRIu64 " %s).\n"),
+                        quotearg_n(0, uc->info->nick),
+                        uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                        quote_n(1, base_name(uc->transfer_file)),
+                        uc->transfer_total - uc->transfer_start,
+                        uc->transfer_total,
+                        ngettext("byte", "bytes", uc->transfer_total));
+            } else {
+	        flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " %s).\n"),
+                        quotearg_n(0, uc->info->nick),
+                        uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                        quote_n(1, base_name(uc->transfer_file)),
+                        uc->transfer_total,
+                        ngettext("byte", "bytes", uc->transfer_total));
+            }
+            uc->transferring = true;
+            uc->transfer_pos = uc->transfer_start;
+            uc->transfer_time = time(NULL);
+            if (uc->transfer_time == (time_t) -1)
+                warn(_("Cannot get current time - %s\n"), errstr);
+            break;
+        case DC_MSG_CHECK_DOWNLOAD:
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
+            for (; uc->queue_pos < uc->info->download_queue->cur; uc->queue_pos++) {
+                DCQueuedFile *queued;
+                char *local_file;
+
+                queued = uc->info->download_queue->buf[uc->queue_pos];
+                if (queued->status != DC_QS_DONE) {
+                    local_file = resolve_download_file(uc->info, queued);
+                    uc->queued_valid = true;
+                    uc->transfer_file = xstrdup(queued->filename);
+                    uc->occupied_slot = true;
+                    queued->status = DC_QS_PROCESSING;
+                    used_dl_slots++;
+                    msgq_put(uc->put_mq, MSGQ_STR, local_file, MSGQ_STR, uc->transfer_file, MSGQ_INT64, queued->length, MSGQ_END);
+                    FD_SET(uc->put_mq->fd, &write_fds);
+                    free(local_file);
+                    return;
+                }
+            }
+            msgq_put(uc->put_mq, MSGQ_STR, NULL, MSGQ_STR, NULL, MSGQ_INT64, (uint64_t) 0, MSGQ_END);
+            FD_SET(uc->put_mq->fd, &write_fds);
+            break;
+        case DC_MSG_DOWNLOAD_ENDED: {
+            bool success;
+            char *reason;
+
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_BOOL, &success, MSGQ_STR, &reason, MSGQ_END);
+            handle_ended_download(uc, success, reason);
+            free(reason);
+            break;
+        }
+        case DC_MSG_CHECK_UPLOAD: {
+            char *remote_file;
+            char *local_file;
+
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_STR, &remote_file, MSGQ_END);
+            if (used_ul_slots >= my_ul_slots && !uc->info->slot_granted) {
+                uc->transfer_file = NULL;
+                uc->occupied_slot = false;
+                local_file = NULL;
+                free(remote_file);
+            } else {
+                uc->transfer_file = remote_file;
+                uc->occupied_slot = true;
+                used_ul_slots++;
+                local_file = resolve_upload_file(uc->info, remote_file);
+            }
+            msgq_put(uc->put_mq, MSGQ_BOOL, uc->occupied_slot, MSGQ_STR, local_file, MSGQ_END);
+            FD_SET(uc->put_mq->fd, &write_fds);
+            free(local_file);
+            break;
+        }
+        case DC_MSG_UPLOAD_ENDED: {
+            bool success;
+            char *reason;
+
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_BOOL, &success, MSGQ_STR, &reason, MSGQ_END);
+            handle_ended_upload(uc, success, reason);
+            free(reason);
+            break;
+        }
+        case DC_MSG_TERMINATING:
+            /* The TERMINATING message will be sent from users when they're about to
+             * shut down properly.
+	     * XXX: This message may contain the reason it was terminated in the future.
+	     */
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
+            user_disconnect(uc); /* MSG: reason from DC_MSG_TERMINATING */
+            return; /* not break! this is the last message. */
+        default:
+            warn(_("Received unknown message %d from user process, shutting down process.\n"), id);
+            user_disconnect(uc); /* MSG: local communication error? */
+            return; /* not break! this is the last message. */
+        }
+    }
+}
 
 void
 transfer_completion_generator(DCCompletionInfo *ci)
@@ -702,6 +765,7 @@ read_signal_input(void)
     }
 
     if (signal == SIGTERM) {
+        warn(_("Received TERM signal, shutting down.\n"));
         running = false;
     }
     else if (signal == SIGINT) {
@@ -718,6 +782,12 @@ read_signal_input(void)
                 screen_wakeup(WIFSIGNALED(status) && WTERMSIG(status) == SIGINT);
                 shell_child = -1;
                 name = _("Shell process");
+            } else if (child == lookup_child) {
+                name = _("Lookup process");
+                running = false;
+            } else if (child == parse_child) {
+                name = _("Parse process");
+                running = false;
             } else {
                 /* Assume it was a child process */
                 name = _("User process");
@@ -731,7 +801,6 @@ read_signal_input(void)
                     sprintf(signame, "%d", WTERMSIG(status));
                 warn(_("%s terminated by signal %s.\n"), name, signame);
             }
-
         }
         if (child < 0 && errno != ECHILD)
             warn(_("Cannot wait for processes - %s\n"), errstr);
@@ -777,6 +846,7 @@ bool
 set_share_dir(const char *dir)
 {
     char *tmp = share_dir;
+    char sizebuf[LONGEST_HUMAN_READABLE+1];
     /* set first, because we may be called as set_share_dir(share_dir); */
     share_dir = (dir == NULL ? NULL : xstrdup(dir));
     free(tmp);
@@ -785,10 +855,10 @@ set_share_dir(const char *dir)
     if (hub_state >= DC_HUB_LOGGED_IN && !send_my_info())
     	return false;
     if (dir != NULL) {
-	screen_putf(_("Sharing %" PRIu64 " %s (%" PRIu64 " MB) totally\n"),
+	screen_putf(_("Sharing %" PRIu64 " %s (%s) totally\n"),
 	    my_share_size,
 	    ngettext("byte", "bytes", my_share_size),
-	    my_share_size/(1024*1024));
+	    human_readable(my_share_size, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
     }
     return true;
 }
@@ -850,7 +920,8 @@ search_now_writable(void)
     	DCUDPMessage *msg = search_udpmsg_out->buf[0];
 	int res;
 
-    	/* <Erwin> I don't think you can send off half a UDP packet, no
+    	/* <Erwin> I don't think you can send off half a UDP packet, no.
+    	 * Therefore we don't attempt to use non-blocking I/O on search_socket.
 	 */
     	res = sendto(search_socket, msg->data, msg->len, 0, (struct sockaddr *) &msg->addr, sizeof(struct sockaddr_in));
         if (res == 0 || (res < 0 && errno != EAGAIN))
@@ -892,7 +963,7 @@ enable_search(void)
     }
 
     if (!fd_set_nonblock_flag(search_socket, true)) {
-    	warn(_("Cannot set socket flags - %s\n"), errstr);
+    	warn(_("Cannot set non-blocking flag - %s\n"), errstr);
 	return false;
     }
 
@@ -933,7 +1004,7 @@ enable_active(uint16_t port)
 
     /* Set non-blocking */
     if (!fd_set_nonblock_flag(listen_socket, true)) {
-    	warn(_("Cannot set socket flags - %s\n"), errstr);
+    	warn(_("Cannot set non-blocking flag - %s\n"), errstr);
 	disable_active();
 	return false;
     }
@@ -1021,10 +1092,10 @@ int
 main (int argc, char **argv)
 {
     uint32_t c;
-    char *filename;
+    char *config_file;
+    bool custom_config;
     char *tmpdir;
     struct sigaction sigact;
-    bool do_config = true;
 
     set_quoting_style(NULL, escape_quoting_style);
     
@@ -1037,14 +1108,23 @@ main (int argc, char **argv)
         warn(_("%s: Cannot set message domain: %s\n"), argv[0], errstr);
 #endif
 
+    custom_config = false;
+    get_package_file("config", &config_file);
+
     while (true) {
 	c = getopt_long(argc, argv, short_opts, long_opts, NULL);
 	if (c == -1)
 	    break;
 
 	switch (c) {
+	case 'c': /* --config */
+	    custom_config = true;
+	    free(config_file);
+	    config_file = xstrdup(optarg);
+	    break;
 	case 'n': /* --no-config */
-	    do_config = false;
+	    free(config_file);
+	    config_file = NULL;
 	    break;
 	case HELP_OPT: /* --help */
 	    printf(_("Usage: %s [OPTION]...\n"), quotearg(argv[0]));
@@ -1130,9 +1210,11 @@ main (int argc, char **argv)
 	warn(_("Cannot find directory for temporary files - %s\n"), errstr);
 	goto cleanup;
     }
-    filename = xasprintf("%s.%d", PACKAGE, getpid());
-    listing_dir = catfiles(tmpdir, filename);
-    free(filename);
+    {
+        char *filename = xasprintf("%s.%d", PACKAGE, getpid());
+        listing_dir = catfiles(tmpdir, filename);
+        free(filename);
+    }
     ptrv_append(delete_dirs, xstrdup(listing_dir));
     is_active = false;
     listen_port = 0;
@@ -1144,11 +1226,16 @@ main (int argc, char **argv)
     	goto cleanup;
     my_ul_slots = 3;
 
+    if (!lookup_init())
+        goto cleanup;
+    if (!file_list_parse_init())
+        goto cleanup;
     command_init();
-    if (do_config) {
-        get_package_file("config", &filename);
-        run_script(filename, true);
-        free(filename);
+
+    if (config_file != NULL) {
+        run_script(config_file, !custom_config);
+        config_file = NULL;
+        free(config_file);
     }      
 
     while (running) {
@@ -1179,13 +1266,24 @@ main (int argc, char **argv)
     	    search_input_available();
 	if (running && search_socket >= 0 && FD_ISSET(search_socket, &res_write_fds))
 	    search_now_writable();
+        if (running && FD_ISSET(lookup_request_mq->fd, &res_write_fds))
+            lookup_request_fd_writable();
+        if (running && FD_ISSET(lookup_result_mq->fd, &res_read_fds))
+            lookup_result_fd_readable();
+        if (running && FD_ISSET(parse_request_mq->fd, &res_write_fds))
+            parse_request_fd_writable();
+        if (running && FD_ISSET(parse_result_mq->fd, &res_read_fds))
+            parse_result_fd_readable();
 	if (running) {
 	    HMapIterator it;
+
 	    hmap_iterator(user_conns, &it);
 	    while (running && it.has_next(&it)) {
 	    	DCUserConn *uc = it.next(&it);
-    	    	if (uc->main_socket >= 0 && FD_ISSET(uc->main_socket, &res_read_fds))
-		    ipc_handle(uc->ipc);
+	    	if (uc->put_mq != NULL && FD_ISSET(uc->put_mq->fd, &res_write_fds))
+	    	    user_request_fd_writable(uc);
+                if (uc->get_mq != NULL && FD_ISSET(uc->get_mq->fd, &res_read_fds))
+                    user_result_fd_readable(uc);
 	    }
 	}
     }
@@ -1195,11 +1293,13 @@ cleanup:
     hub_disconnect();
     screen_finish();
     command_finish();
+    file_list_parse_finish();
+    lookup_finish();
 
     byteq_free(hub_recvq);
     byteq_free(hub_sendq);
-    hmap_free(hub_users); /* Empied by hub_disconnect */
-    hmap_free(pending_userinfo); /* Empied by hub_disconnect */
+    hmap_free(hub_users); /* Emptied by hub_disconnect */
+    hmap_free(pending_userinfo); /* Emptied by hub_disconnect */
 
     byteq_free(search_recvq);
 
@@ -1213,11 +1313,13 @@ cleanup:
     ptrv_free(our_searches);
 
     hmap_foreach_value(user_conns, user_conn_cancel);
+    /* XXX: follow up and wait for user connections to die? */
     hmap_free(user_conns);
 
     if (our_filelist != NULL)
     	filelist_free(our_filelist);
 
+    free(hub_charset);
     free(hub_name);
     free(my_nick);
     free(my_description);
@@ -1227,6 +1329,11 @@ cleanup:
     free(download_dir);
     free(listing_dir);
 
+    if (to_hub_iconv != (iconv_t) -1 && iconv_close(to_hub_iconv) != 0)
+        warn(_("cannot close character convertor - %s\n"), errstr);
+    if (from_hub_iconv != (iconv_t) -1 && iconv_close(from_hub_iconv) != 0)
+        warn(_("cannot close character convertor - %s\n"), errstr);
+
     if (delete_files != NULL) {
 	for (c = 0; c < delete_files->cur; c++) {
     	    char *filename = delete_files->buf[c];
@@ -1272,5 +1379,7 @@ cleanup:
     if (signal_pipe[1] >= 0 && close(signal_pipe[1]) < 0)
     	warn(_("Cannot close signal pipe - %s\n"), errstr);
 
+    free(config_file);
+
     exit(EXIT_SUCCESS);
 }
