@@ -25,12 +25,15 @@
 #include <stdint.h>
 #include <netinet/in.h>
 #include <time.h>
+#include <netdb.h>		/* struct addrinfo */
 #include <fcntl.h>		/* POSIX: O_NONBLOCK */
+#include <dirent.h>		/* POSIX: DIR, struct dirent */
+#include <iconv.h>		/* Gnulib, GNU libc: iconv_t */
 #include "common/byteq.h"
 #include "common/ptrv.h"
 #include "common/error.h"
 #include "common/hmap.h"
-#include "ipc.h"
+#include "common/msgq.h"
 
 #define DC_CLIENT_BASE_KEY 5
 #define DC_HUB_TCP_PORT 411
@@ -92,6 +95,7 @@ typedef enum {
 
 typedef enum {
     DC_HUB_DISCONNECTED,
+    DC_HUB_LOOKUP,		/* Waiting for getaddrinfo lookup to complete */
     DC_HUB_CONNECT,	    	/* Waiting for connect() to complete or socket to be writable */
     DC_HUB_LOCK,	    	/* Waiting for $Lock */
     DC_HUB_HELLO,		/* Waiting for $Hello */
@@ -112,13 +116,13 @@ typedef enum {
     DC_MSG_VALIDATE_DIR,
     DC_MSG_VALIDATE_NICK,
     DC_MSG_GET_MY_NICK,
-    DC_MSG_DISCONNECT,
     DC_MSG_CHECK_DOWNLOAD,	/* get information on next download, allocate slot. */
     DC_MSG_CHECK_UPLOAD,	/* check that upload is allowed, allocate slot. */
-    DC_MSG_UPLOAD_ENDED,	/* free slot. */
+    DC_MSG_UPLOAD_ENDED,	/* free slot and print info about upload. */
     DC_MSG_DOWNLOAD_ENDED,	/* free slot and mark download as done or failed. */
     DC_MSG_TRANSFER_START,
     DC_MSG_TRANSFER_STATUS,
+    DC_MSG_TERMINATING,		/* the user process decided to terminate. */
 } DCUserMsgId;
 
 typedef enum {
@@ -152,10 +156,15 @@ typedef struct _DCUDPMessage DCUDPMessage;
 typedef struct _DCSearchResponse DCSearchResponse;
 typedef struct _DCQueuedFile DCQueuedFile;
 typedef struct _DCVariable DCVariable;
+typedef struct _DCLookup DCLookup; /* defined in lookup.c */
+typedef struct _DCFileListParse DCFileListParse; /* defined in filelist-in.c */
 
 typedef void (*DCCompletorFunction)(DCCompletionInfo *ci);
 typedef void (*DCBuiltinCommandHandler)(int argc, char **argv);
 typedef void (*ScreenWriter)(DCDisplayFlag flag, const char *format, va_list args);
+typedef void (*DCLookupCallback)(int rc, struct addrinfo *result_ai, void *data);
+/* This callback is responsible for freeing node when no longer needed. */
+typedef void (*DCFileListParseCallback)(DCFileList *node, void *data);
 
 struct _DCSearchResponse {
     uint32_t refcount;
@@ -256,16 +265,15 @@ struct _DCFileList {
     DCFileList *parent;
     char *name;
     DCFileType type;
+    uint64_t size;	/* total size of all contained files for DC_TYPE_DIR */
     union {
     	struct {
-    	    uint64_t size;
-	    char tth[39];
+	    /*char tth[39];*/
 	} reg;
 	struct {
-    	    uint64_t totalsize;
 	    HMap *children;
 	} dir;
-    } u;
+    };
 };
 
 struct _DCQueuedFile {
@@ -296,11 +304,16 @@ struct _DCUserInfo {
 
 struct _DCUserConn {
     char *name;     	/* key of connection */
+    bool disconnecting;
     DCUserInfo *info;	/* if info!=NULL then info->conn must contain this DCUserConn. */
     DCTransferDirection dir;
     pid_t pid;
-    int main_socket;
-    IPC *ipc;
+    //int main_socket;
+    //IPC *ipc;
+    MsgQ *get_mq;
+    MsgQ *put_mq;	/* Note: The main process may only put message on put_mq as a result of
+                         * a message just received on get_mq.
+                         */
     bool occupied_slot; /* true if used_*_slots were increased for this user connection */
     uint32_t queue_pos;
     bool queued_valid;		/* true unless the QueuedFile by queue_pos has been removed */
@@ -330,6 +343,9 @@ extern char *log_filename;
 extern DCHubState hub_state;
 extern ByteQ *hub_recvq;
 extern ByteQ *hub_sendq;
+extern char *hub_charset;
+extern iconv_t to_hub_iconv;
+extern iconv_t from_hub_iconv;
 extern HMap *hub_users;
 extern HMap *user_conns;
 extern struct sockaddr_in local_addr;
@@ -375,6 +391,7 @@ bool hub_connect_user(DCUserInfo *ui);
 DCUserInfo *user_info_new(const char *nick);
 DCUserInfo *find_hub_user(const char *nick);
 void user_info_free(DCUserInfo *ui);
+void hub_new(const char *hostname, uint16_t port);
 void hub_connect(struct sockaddr_in *addr);
 void hub_disconnect(void);
 void hub_now_writable(void);
@@ -392,6 +409,7 @@ void default_completion_selector(DCCompletionInfo *ci);
 void update_prompt(void);
 void command_init(void);
 void command_finish(void);
+void browse_list_parsed(DCFileList *node, void *data);
 
 /* screen.c */
 #define screen_putf(f,...) flag_putf(DC_DF_COMMON, (f), ## __VA_ARGS__)
@@ -422,10 +440,9 @@ char *huffman_decode(const uint8_t *data, uint32_t data_size, uint32_t *out_size
 char *huffman_encode(const uint8_t *data, uint32_t data_size, uint32_t *out_size);
 
 /* user.c */
-void user_main(int main_sockets[2], struct sockaddr_in *addr, int sock);
+void user_main(int get_fd[2], int put_fd[2], struct sockaddr_in *addr, int sock);
 
 /* main.c */
-void add_user_conn(DCUserConn *uc);
 /*bool get_user_conn_status(DCUserConn *uc);*/
 bool set_active(bool newactive, uint16_t newport);
 bool set_share_dir(const char *dir);
@@ -435,20 +452,20 @@ bool get_package_file(const char *name, char **outname);
 void transfer_completion_generator(DCCompletionInfo *ci);
 void user_conn_cancel(DCUserConn *uc);
 void warn_file_error(int res, bool write, const char *filename);
-void warn_socket_error(int res, bool write, const char *subject);
+void warn_socket_error(int res, bool write, const char *subject, ...);
 void add_search_result(struct sockaddr_in *addr, char *results, uint32_t resultlen);
 void free_queued_file(DCQueuedFile *qf); /* XXX: move transfer.c? */
 DCUserConn *user_connection_new(struct sockaddr_in *addr, int socket);
-void user_disconnect(DCUserConn *uc);
+/*void user_disconnect(DCUserConn *uc);*/
 char *user_conn_status_to_string(DCUserConn *uc, time_t now);
 extern uint64_t bytes_received;
 extern uint64_t bytes_sent;
 
 /* fs.c */
+DCFileList *new_file_node(const char *name, DCFileType type, DCFileList *parent);
 void filelist_free(DCFileList *fl);
 void filelist_list(DCFileList *fl, bool longmode);
 DCFileList *filelist_lookup(DCFileList *node, const char *filename);
-DCFileList *filelist_open(const char *filename);
 char *filelist_get_path(DCFileList *node);
 char *filelist_get_path_with_trailing_slash(DCFileList *node);
 char *resolve_upload_file(DCUserInfo *ui, const char *name);
@@ -469,7 +486,8 @@ bool has_leading_slash(const char *str);
     
 /* connection.c */
 char *decode_lock(const char *lock, size_t locklen, uint32_t basekey);
-char *escape_message(char *str);
+char *escape_message(const char *str);
+char *unescape_message(const char *str);
 void dump_command(const char *header, const char *buf, size_t len);
 
 /* util.c */
@@ -493,6 +511,10 @@ bool parse_ip_and_port(char *source, struct sockaddr_in *addr, uint16_t defport)
 #define quotearg_mem(m,l)	(quotearg_n_style_mem(0, escape_quoting_style, (m), (l)))
 char *join_strings(char **strs, int count, char mid);
 PtrV *wordwrap(const char *str, size_t len, size_t first_width, size_t other_width);
+struct dirent *xreaddir(DIR *dh);
+    
+#define LONGEST_ELAPSED_TIME 22 /* 123456789012dNNhNNmNNs */
+char *elapsed_time_to_string(time_t elapsed, char *buf);
 
 /* search.c */
 bool parse_search_selection(char *str, DCSearchSelection *data);
@@ -507,4 +529,26 @@ char *search_selection_to_string(DCSearchSelection *sr);
 void cmd_set(int argc, char **argv);
 void set_command_completion_selector(DCCompletionInfo *ci);
 
+/* lookup.c */
+extern MsgQ *lookup_request_mq;
+extern MsgQ *lookup_result_mq;
+extern pid_t lookup_child;
+bool lookup_init(void);
+void lookup_request_fd_writable(void);
+void lookup_result_fd_readable(void);
+DCLookup *add_lookup_request(const char *node, const char *service, const struct addrinfo *hints, DCLookupCallback callback, void *userdata);
+void cancel_lookup_request(DCLookup *lookup);
+void lookup_finish(void);
+
+/* filelist-in.c */
+extern MsgQ *parse_request_mq;
+extern MsgQ *parse_result_mq;
+extern pid_t parse_child;
+bool file_list_parse_init(void);
+void file_list_parse_finish(void);
+DCFileListParse *add_parse_request(DCFileListParseCallback callback, const char *filename, void *userdata);
+void cancel_parse_request(DCFileListParse *parse);
+void parse_result_fd_readable(void);
+void parse_request_fd_writable(void);
+
 #endif
