@@ -243,7 +243,7 @@ append_result(DCFileList *node, DCUserInfo *ui, struct sockaddr_in *addr)
 
     strbuf_appendf(sb, "$SR %s %s", my_nick, rpath);
     if (node->type == DC_TYPE_REG)
-    	strbuf_appendf(sb, "\x05%" PRIu64, node->u.reg.size);
+    	strbuf_appendf(sb, "\x05%" PRIu64, node->size);
     strbuf_appendf(sb, " %d/%d\x05%s (%s)",
     	    free_slots, my_ul_slots, hub_name,
 	    sockaddr_in_str(&hub_addr));
@@ -268,9 +268,9 @@ filelist_search(DCFileList *node, DCSearchSelection *data, int maxresults, DCUse
     if (node->type == DC_TYPE_REG) {
     	if (data->datatype == DC_SEARCH_FOLDERS || data->datatype == DC_SEARCH_CHECKSUM) /* TTH not supported yet */
 	    return 0;
-    	if (node->u.reg.size < data->size_min)
+    	if (node->size < data->size_min)
 	    return 0;
-	if (node->u.reg.size > data->size_max)
+	if (node->size > data->size_max)
 	    return 0;
 	if (!match_search_patterns(node->name, data))
 	    return 0;
@@ -293,7 +293,7 @@ filelist_search(DCFileList *node, DCSearchSelection *data, int maxresults, DCUse
 	    }
 	}
 
-	hmap_iterator(node->u.dir.children, &it);
+	hmap_iterator(node->dir.children, &it);
 	while (it.has_next(&it)) {
     	    DCFileList *subnode = it.next(&it);
 	    curresults += filelist_search(subnode, data, maxresults-curresults, ui, addr);
@@ -319,9 +319,9 @@ perform_inbound_search(DCSearchSelection *data, DCUserInfo *ui, struct sockaddr_
 
     if (curresults > 0) {
         if (ui != NULL) {
-            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s\n"), curresults, maxresults, quotearg(ui->nick));
+            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s.\n"), curresults, maxresults, quotearg(ui->nick));
         } else {
-            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s\n"), curresults, maxresults, sockaddr_in_str(addr));
+            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s.\n"), curresults, maxresults, sockaddr_in_str(addr));
         }
     } else {
         flag_putf(DC_DF_DEBUG, _("No search results.\n"));
