@@ -53,10 +53,13 @@
 #define DEFAULT_RECVQ_SIZE (64*1024)
 #define DEFAULT_SENDQ_SIZE (64*1024)
 
+#define USER_CONN_IDLE_TIMEOUT (3*60)
+
 typedef struct _DCUserConnLocal DCUserConnLocal;
 
 struct _DCUserConnLocal {
-    int main_socket/* = -1*/;
+    MsgQ *get_mq;/*=NULL*/
+    MsgQ *put_mq;/*=NULL*/
     int signal_pipe[2]/* = { -1, -1 }*/;
     char *user_nick/* = NULL*/;
     uint32_t user_recvq_last/* = 0*/;
@@ -71,7 +74,6 @@ struct _DCUserConnLocal {
     bool user_running/* = true*/;
     fd_set user_read_fds;
     fd_set user_write_fds;
-    IPC *ipc;
 
     //union {
       //struct {
@@ -101,12 +103,51 @@ static DCUserConnLocal *cur_ucl;
  * things happen gracefully.
  */
 
+/* This function should be called when the communication to the
+ * main process becomes unusable. It will terminate the user
+ * process. If it appears that the communication was closed by
+ * the remote end, no error is printed.
+ */
+static void
+fatal_error(DCUserConnLocal *ucl, int res, bool writing) /* XXX: rename communication_error or something */
+{
+    warn_writer = default_warn_writer;
+    /* These tests are for the case when the main process closed
+     * the connection first. The first test is when reading (and
+     * the result is EOF because of closed pipe), and the second
+     * set of tests are used when writing. The latter case is
+     * probably rare - the child process would have to be
+     * 'cancelled' in main when the child was just about to call
+     * msgq_put.
+     */
+    if (res != 0 && !(writing && res < 0 && errno == EPIPE))
+        warn_socket_error(res, writing, _("main process"));
+    ucl->user_running = false;
+}
+
+/* This function should be called when the user process decided to
+ * terminate itself (but the communication to the main process
+ * is still usable).
+ */
+static void
+terminate_process(DCUserConnLocal *ucl)
+{
+    int res;
+
+    ucl->user_running = false;
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_TERMINATING, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return;
+    }
+}
+
 static bool
 check_state(DCUserConnLocal *ucl, char *buf, DCUserState state)
 {
     if (ucl->user_state != state) {
 	warn(_("Received %s message in wrong state.\n"), strtok(buf, " "));
-	ucl->user_running = false;
+	terminate_process(ucl); /* protocol error */
 	return false;
     }
     return true;
@@ -115,45 +156,14 @@ check_state(DCUserConnLocal *ucl, char *buf, DCUserState state)
 static bool
 send_user_status(DCUserConnLocal *ucl, uint64_t pos)
 {
-    return ipc_put(ucl->ipc, DC_MSG_TRANSFER_STATUS, IPC_INT64, pos, IPC_END);
-}
+    int res;
 
-static bool
-handle_main_msg(IPC *ipc, uint32_t id, void *userdata)
-{
-    DCUserConnLocal *ucl = userdata;
-
-    switch (id) {
-    case DC_MSG_DISCONNECT:
-    	ipc_get(ipc, id, IPC_END);
-	ipc_put(ipc, id, IPC_END); /* Ignore errors */
-    	ucl->user_running = false;
-	return false;
-    default:
-        warn(_("Received unknown message %d from main process, shutting down.\n"), id);
-	ucl->user_running = false;
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_TRANSFER_STATUS, MSGQ_INT64, pos, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
         return false;
     }
-}
-
-static void
-handle_main_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
-{
-    DCUserConnLocal *ucl = userdata;
-
-    warn_writer = default_warn_writer;
-    switch (error) {
-    case IPC_ERROR_WRITE:
-    	warn_socket_error(result, true, _("main process"));
-	break;
-    case IPC_ERROR_READ:
-    	warn_socket_error(result, false, _("main process"));
-	break;
-    case IPC_ERROR_DATA:
-    	warn(_("Logical communication error with main process (during message %d).\n"), result);
-    	break;
-    }
-    ucl->user_running = false;
+    return true;
 }
 
 static void
@@ -161,10 +171,13 @@ user_screen_writer(DCDisplayFlag flag, const char *format, va_list args)
 {
     DCUserConnLocal *ucl = cur_ucl; /* XXX */
     char *msg;
+    int res;
 
     msg = xvasprintf(format, args);
-    ipc_put(ucl->ipc, DC_MSG_SCREEN_PUT, IPC_INT32, (int32_t) flag, IPC_STR, msg, IPC_END); /* ignore errors */
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_SCREEN_PUT, MSGQ_INT32, (int32_t) flag, MSGQ_STR, msg, MSGQ_END);
     free(msg);
+    if (res <= 0) /* XXX: should we maybe print this message manually? */
+        fatal_error(ucl, res, true); /* delay handling until later. */
 }
 
 /* Put some data onto the connection, printf style.
@@ -181,11 +194,12 @@ user_putf(DCUserConnLocal *ucl, const char *format, ...)
     res = byteq_vappendf(ucl->user_sendq, format, args);
     va_end(args);
 
-    if (res < 0) {
+    /* byteq_vappendf cannot fail. */
+    /*if (res < 0) {
     	warn(_("Cannot append to user send queue - %s\n"), errstr);
-	ucl->user_running = false;
+	terminate_process(ucl);
 	return false;
-    }
+    }*/
 
     if (ucl->data_size/*DL*/ == 0)
         dump_command(_("-->"), ucl->user_sendq->buf+oldcur, ucl->user_sendq->cur-oldcur);
@@ -193,7 +207,7 @@ user_putf(DCUserConnLocal *ucl, const char *format, ...)
     res = byteq_write(ucl->user_sendq, ucl->user_socket);
     if (res == 0 || (res < 0 && errno != EAGAIN)) {
 	warn_socket_error(res, true, _("user"));
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: socket error above */
 	return false;
     }
 
@@ -206,10 +220,18 @@ user_putf(DCUserConnLocal *ucl, const char *format, ...)
 static bool
 wants_to_download(DCUserConnLocal *ucl, bool *reply)
 {
-    if (!ipc_put(ucl->ipc, DC_MSG_WANT_DOWNLOAD, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_WANT_DOWNLOAD, IPC_BOOL, reply, IPC_END))
-    	return false;
+    int res;
+
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_WANT_DOWNLOAD, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return false;
+    }
+    res = msgq_get_sync(ucl->get_mq, MSGQ_BOOL, reply, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
+        return false;
+    }
     return true;
 }
 
@@ -217,14 +239,21 @@ static bool
 direction_validate(DCUserConnLocal *ucl, DCTransferDirection dir)
 {
     bool reply;
+    int res;
 
-    if (!ipc_put(ucl->ipc, DC_MSG_VALIDATE_DIR, IPC_INT, dir, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_VALIDATE_DIR, IPC_BOOL, &reply, IPC_END))
-    	return false;
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_VALIDATE_DIR, MSGQ_INT, dir, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return false;
+    }
+    res = msgq_get_sync(ucl->get_mq, MSGQ_BOOL, &reply, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
+        return false;
+    }
     if (!reply) {
 	warn(_("Too many connections to user, or no free slots.\n"));
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: msg above */
 	return false;
     }
     return true;
@@ -233,10 +262,18 @@ direction_validate(DCUserConnLocal *ucl, DCTransferDirection dir)
 static bool
 get_our_nick(DCUserConnLocal *ucl, char **str)
 {
-    if (!ipc_put(ucl->ipc, DC_MSG_GET_MY_NICK, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_GET_MY_NICK, IPC_STR, str, IPC_END))
-    	return false;
+    int res;
+
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_GET_MY_NICK, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return false;
+    }
+    res = msgq_get_sync(ucl->get_mq, MSGQ_STR, str, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
+        return false;
+    }
     return true;
 }
 
@@ -244,14 +281,21 @@ static bool
 nick_validate(DCUserConnLocal *ucl, const char *str)
 {
     bool reply;
+    int res;
 
-    if (!ipc_put(ucl->ipc, DC_MSG_VALIDATE_NICK, IPC_STR, str, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_VALIDATE_NICK, IPC_BOOL, &reply, IPC_END))
-    	return false;
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_VALIDATE_NICK, MSGQ_STR, str, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return false;
+    }
+    res = msgq_get_sync(ucl->get_mq, MSGQ_BOOL, &reply, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
+        return false;
+    }
     if (!reply) {
 	warn(_("User %s not on hub, or too many connections to user.\n"), quotearg(str));/*XXX: move main.c */
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: msg above */
     	return false;
     }
     return true;
@@ -262,6 +306,7 @@ end_download(DCUserConnLocal *ucl, bool success, const char *reason_fmt, ...)
 {
     va_list args;
     char *reason = NULL;
+    int res;
 
     if (ucl->transfer_fd/*DL*/ >= 0) {
         /* Inability to close the downloaded file is a serious error. */
@@ -278,7 +323,9 @@ end_download(DCUserConnLocal *ucl, bool success, const char *reason_fmt, ...)
     va_start(args, reason_fmt);
     if (reason == NULL)
         reason = xvasprintf(reason_fmt, args);
-    ipc_put(ucl->ipc, DC_MSG_DOWNLOAD_ENDED, IPC_BOOL, success, IPC_STR, reason, IPC_END);
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_DOWNLOAD_ENDED, MSGQ_BOOL, success, MSGQ_STR, reason, MSGQ_END);
+    if (res <= 0)
+        fatal_error(ucl, res, true);
     free(reason);
     va_end(args);
 
@@ -300,16 +347,23 @@ download_next_file(DCUserConnLocal *ucl)
     uint64_t resume_pos;
     uint64_t file_size;
     struct stat sb;
+    int res;
 
     if (!ucl->user_running)
         return;
-    if (!ipc_put(ucl->ipc, DC_MSG_CHECK_DOWNLOAD, IPC_END))
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_CHECK_DOWNLOAD, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
     	return;
-    if (!ipc_get(ucl->ipc, DC_MSG_CHECK_DOWNLOAD, IPC_STR, &local_file, IPC_STR, &share_file, IPC_INT64, &file_size, IPC_END))
+    }
+    res = msgq_get_sync(ucl->get_mq, MSGQ_STR, &local_file, MSGQ_STR, &share_file, MSGQ_INT64, &file_size, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
     	return;
+    }
     if (local_file == NULL) {
         flag_putf(DC_DF_CONNECTIONS, _("No more files to download.\n"));/*XXX:move main.c?*/
-        ucl->user_running = false;
+	terminate_process(ucl); /* MSG: no more to download */
 	return;
     }
 
@@ -321,7 +375,7 @@ download_next_file(DCUserConnLocal *ucl)
     if (lstat(local_file, &sb) != 0) {
         if (errno != ENOENT) {
             warn(_("%s: Cannot get file status: %s\n"), quotearg(local_file), errstr);
-            end_download(ucl, false, "local error");
+            end_download(ucl, false, _("local error"));
             download_next_file(ucl);
             return;
         }
@@ -330,7 +384,7 @@ download_next_file(DCUserConnLocal *ucl)
     } else {
         if (!S_ISREG(sb.st_mode)) {
             warn(_("%s: File exists and is not a regular file\n"), quotearg(local_file));
-            end_download(ucl, false, "local error");
+            end_download(ucl, false, _("local error"));
             download_next_file(ucl);
             return;
         }
@@ -340,7 +394,7 @@ download_next_file(DCUserConnLocal *ucl)
 
     remote_file = translate_local_to_remote(share_file);
     if (!user_putf(ucl, "$Get %s$%" PRIu64 "|", remote_file, resume_pos+1)) { /* " */
-        end_download(ucl, false, "communication error");
+        end_download(ucl, false, _("communication error"));
         free(remote_file);
 	return;
     }
@@ -354,6 +408,8 @@ download_next_file(DCUserConnLocal *ucl)
 static void
 open_download_file(DCUserConnLocal *ucl, uint64_t file_size)
 {
+    int res;
+
     if (ucl->file_size == UINT64_MAX) {
         /* A file size of UINT64_MAX means that we did not know the size
          * of the file in advance which is only true for file list files.
@@ -361,7 +417,7 @@ open_download_file(DCUserConnLocal *ucl, uint64_t file_size)
         ucl->file_size = file_size;
     } else if (file_size < ucl->file_size) {
         end_download(ucl, false,
-            "remote file is smaller than local (expected %" PRIu64 ", got %" PRIu64 " %s)",
+            _("remote file is smaller than local (expected %" PRIu64 ", got %" PRIu64 " %s)"),
             ucl->file_size, file_size, ngettext("byte", "bytes", file_size));
         /* We're probably left in a weird state here. It is possible that
          * some clients (at least DC++ 0.700) won't let us download more
@@ -378,30 +434,32 @@ open_download_file(DCUserConnLocal *ucl, uint64_t file_size)
     }
     if (ucl->transfer_fd/*DL*/ < 0) {
         warn(_("%s: Cannot open file for writing - %s\n"), quotearg(ucl->local_file/*DL*/), errstr);
-        end_download(ucl, false, "local error");
+        end_download(ucl, false, _("local error"));
         download_next_file(ucl);
         return;
     }
     if (ucl->file_pos != 0 && lseek(ucl->transfer_fd/*DL*/, ucl->file_pos, SEEK_SET) < 0) {
         warn(_("%s: Cannot seek to resume position - %s\n"), quotearg(ucl->local_file/*DL*/), errstr);
-        end_download(ucl, false, "local error");
+        end_download(ucl, false, _("local error"));
         download_next_file(ucl);
         return;
     }
 
     if (!user_putf(ucl, "$Send|")) {
-        end_download(ucl, false, "communication error");
+        end_download(ucl, false, _("communication error"));
         return;
     }
 
-    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_START, IPC_INT64, ucl->file_pos, IPC_INT64, ucl->file_size, IPC_END)) {
-        end_download(ucl, false, "internal communication error");
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_TRANSFER_START, MSGQ_INT64, ucl->file_pos, MSGQ_INT64, ucl->file_size, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        /*end_download(ucl, false, "internal communication error");*/ /* Just won't go through! */
 	return;
     }
 
     ucl->data_size/*DL*/ = ucl->file_size - ucl->file_pos;
     if (ucl->data_size/*DL*/ == 0) {
-        end_download(ucl, true, "no data to transfer");
+        end_download(ucl, true, _("no data to transfer"));
         download_next_file(ucl);
         return;
     }
@@ -417,8 +475,10 @@ open_download_file(DCUserConnLocal *ucl, uint64_t file_size)
 }
 
 static void
-end_upload(DCUserConnLocal *ucl, bool success)
+end_upload(DCUserConnLocal *ucl, bool success, const char *reason)
 {
+    int res;
+
     if (ucl->transfer_fd/*DL*/ >= 0) {
         close(ucl->transfer_fd/*DL*/);  /* Ignore errors */
         ucl->transfer_fd/*DL*/ = -1;
@@ -427,7 +487,9 @@ end_upload(DCUserConnLocal *ucl, bool success)
     ucl->share_file/*UL*/ = NULL;
     free(ucl->local_file/*UL*/);
     ucl->local_file/*UL*/ = NULL;
-    ipc_put(ucl->ipc, DC_MSG_UPLOAD_ENDED, IPC_BOOL, success, IPC_END);
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_UPLOAD_ENDED, MSGQ_BOOL, success, MSGQ_STR, reason, MSGQ_END);
+    if (res <= 0)
+        fatal_error(ucl, res, true);
 }
 
 static void
@@ -437,13 +499,16 @@ open_upload_file(DCUserConnLocal *ucl, const char *str, uint64_t offset)
     bool may_upload;
     char *share_file;
     char *local_file;
+    int res;
 
     share_file = translate_remote_to_local(str);
-    if (!ipc_put(ucl->ipc, DC_MSG_CHECK_UPLOAD, IPC_STR, share_file, IPC_END)) {
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_CHECK_UPLOAD, MSGQ_STR, share_file, MSGQ_END);
+    if (res <= 0) {
         free(share_file);
         return;
     }
-    if (!ipc_get(ucl->ipc, DC_MSG_CHECK_UPLOAD, IPC_BOOL, &may_upload, IPC_STR, &local_file, IPC_END)) {
+    res = msgq_get_sync(ucl->get_mq, MSGQ_BOOL, &may_upload, MSGQ_STR, &local_file, MSGQ_END);
+    if (res <= 0) {
         free(share_file);
         return;
     }
@@ -459,37 +524,37 @@ open_upload_file(DCUserConnLocal *ucl, const char *str, uint64_t offset)
     if (ucl->local_file/*UL*/ == NULL) {
         flag_putf(DC_DF_CONNECTIONS, _("%s: No such shared file\n"), quotearg(ucl->share_file/*UL*/));
 	user_putf(ucl, "$Error File Not Available|");
-	end_upload(ucl, false);
+	end_upload(ucl, false, _("no such shared file"));
 	return;
     }
     if (stat(ucl->local_file/*UL*/, &st) < 0) {
         flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot get file status - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
         user_putf(ucl, "$Error File Not Available|");
-        end_upload(ucl, false);
+        end_upload(ucl, false, _("local error"));
         return;
     }
     if (offset > st.st_size) {
 	flag_putf(DC_DF_CONNECTIONS, _("%s: Resume offset %" PRIu64 " outside file\n"), quotearg(ucl->local_file/*UL*/), offset);
 	user_putf(ucl, "$Error Offset out of range|");
-	end_upload(ucl, false);
+	end_upload(ucl, false, _("resume offset out of range"));
 	return;
     }
     ucl->transfer_fd/*UL*/ = open/*64*/(ucl->local_file/*UL*/, O_RDONLY);
     if (ucl->transfer_fd/*UL*/ < 0) {
 	flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot open file for reading - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
 	user_putf(ucl, "$Error File Not Available|");
-	end_upload(ucl, false);
+	end_upload(ucl, false, _("local error"));
 	return;
     }
     if (offset != 0 && lseek/*64*/(ucl->transfer_fd/*UL*/, offset, SEEK_SET) < 0) {
     	flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot seek in file - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
 	user_putf(ucl, "$Error File Not Available|");
-	end_upload(ucl, false);
+	end_upload(ucl, false, _("local error"));
 	return;
     }
     ucl->file_size = st.st_size;
     if (!user_putf(ucl, "$FileLength %llu|", ucl->file_size)) {
-        end_upload(ucl, false);
+        end_upload(ucl, false, _("communication error"));
         return;
     }
     ucl->file_pos = offset;
@@ -512,14 +577,18 @@ open_upload_file(DCUserConnLocal *ucl, const char *str, uint64_t offset)
 static void
 upload_file(DCUserConnLocal *ucl)
 {
+    int res;
+
     if (ucl->file_size == ucl->file_pos) {
         /* There's nothing to send - size is 0 or resuming at end. */
-        end_upload(ucl, true);
+        end_upload(ucl, true, _("no data to transfer"));
         ucl->user_state = DC_USER_GET;
         return;
     }
-    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_START, IPC_INT64, ucl->file_pos, IPC_INT64, ucl->file_size, IPC_END)) {
-        end_upload(ucl, true);
+    res = msgq_put_sync(ucl->put_mq, MSGQ_INT, DC_MSG_TRANSFER_START, MSGQ_INT64, ucl->file_pos, MSGQ_INT64, ucl->file_size, MSGQ_END);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        /*end_upload(ucl, true);*/ /* Just won't go through */
         return;
     }
     FD_SET(ucl->user_socket, &ucl->user_write_fds);
@@ -539,8 +608,8 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	if (res < len) {
 	    /* We cannot expect to synchronize with remote at this point, so shut down. */
 	    warn_file_error(res, true, ucl->share_file/*DL*/);
-	    end_download(ucl, false, "local error");
-	    ucl->user_running = false; 
+	    end_download(ucl, false, _("local error"));
+	    terminate_process(ucl); /* MSG: local error */
 	    return;
 	}
 	ucl->file_pos += len;
@@ -548,7 +617,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	ucl->data_size/*DL*/ -= len;
     	send_user_status(ucl, ucl->file_pos);
     	if (ucl->file_pos == ucl->file_size) {
-	    end_download(ucl, true, "transfer complete");
+	    end_download(ucl, true, _("transfer complete"));
 	    download_next_file(ucl);
 	    return;
 	}
@@ -611,7 +680,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
     	token = strtok(buf+11, " ");
 	if (token == NULL) {
 	    warn(_("Invalid $Direction message: Missing direction parameter\n"));
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: protocol error */
 	    return;
 	}
     	if (strcmp(token, "Upload") == 0) {
@@ -620,19 +689,19 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	    they_download = true;
 	} else {
 	    warn(_("Invalid $Direction message: Invalid direction parameter\n"));
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: protocol error */
 	    return;
 	}
 
 	token = strtok(NULL, " ");
 	if (token == NULL) {
 	    warn(_("Invalid $Direction message: Missing challenge parameter\n"));
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: protocol error */
 	    return;
 	}
 	if (!parse_uint16(token, &remote_rand)) {
 	    warn(_("Invalid $Direction message: Invalid challenge parameter\n"));
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: protocol error */
 	    return;
 	}
 
@@ -653,7 +722,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	} else { /* Remote wants to upload. Do we want to upload too? */
 	    if (!we_download) {
 		warn(_("User does not want to download, nor do we.\n"));
-		ucl->user_running = false;
+		terminate_process(ucl); /* MSG: no one wants to exchange files */
 		return;
 	    }
 	    ucl->our_dir = DC_DIR_RECEIVE;
@@ -683,7 +752,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 
 	if (ucl->user_state != DC_USER_SEND_GET && ucl->user_state != DC_USER_GET) {
 	    warn(_("Received %s message in wrong state.\n"), strtok(buf, " "));
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: protocol error */
 	    return;
 	}
 	token = strtok(buf+5, "$");
@@ -694,7 +763,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	    token = strtok(NULL, ""); /* Cannot fail! */
 	    if (!parse_uint64(token, &offset)) {
 	    	warn(_("Invalid $Get message: Offset not integer\n"));
-		ucl->user_running = false;
+		terminate_process(ucl); /* MSG: protocol error */
 		return;
 	    }
 	    if (offset > 0)
@@ -703,15 +772,15 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 
         /* Maybe the remote user did not want the file after all? */
 	if (ucl->user_state == DC_USER_SEND_GET)
-	    end_upload(ucl, false);
+	    end_upload(ucl, false, _("remote did not want file"));
 
     	open_upload_file(ucl, buf+5, offset); /* Changes state */
     }
     else if (len == 9 && strcmp(buf, "$MaxedOut") == 0) {
     	if (!check_state(ucl, buf, DC_USER_FILE_LENGTH))
 	    return;
-	end_download(ucl, false, "remote is maxed out");
-	ucl->user_running = false;
+	end_download(ucl, false, _("remote is maxed out"));
+	terminate_process(ucl); /* MSG: msg above */
     }
     else if (len >= 12 && strncmp(buf, "$FileLength ", 12) == 0) {
         uint64_t file_size;
@@ -719,7 +788,7 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
     	if (!check_state(ucl, buf, DC_USER_FILE_LENGTH))
 	    return;
 	if (!parse_uint64(buf+12, &file_size)) {
-	    end_download(ucl, false, "protocol error: invalid $FileLength message");
+	    end_download(ucl, false, _("protocol error: invalid $FileLength message"));
 	    download_next_file(ucl);
 	    return;
 	}
@@ -728,15 +797,15 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
     else if (len >= 7 && strncmp(buf, "$Error ", 7) == 0) {
 	if (ucl->user_state == DC_USER_FILE_LENGTH) {
     	    if (strcmp(buf+7, "File Not Available") == 0) {
-	    	end_download(ucl, false, "file not available on remote");
+	    	end_download(ucl, false, _("file not available on remote"));
 	    } else {
-	        end_download(ucl, false, "remote error: %s", quotearg(buf+7));
+	        end_download(ucl, false, _("remote error: %s"), quotearg(buf+7));
 	    }
             download_next_file(ucl);
 	    return;
 	}
 	warn(_("Received error from user: %s\n"), quotearg(buf+7));
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: remote replied with error (including msg above) */
     }
     else if (len >= 5 && strncmp(buf, "$Send ", 5) == 0) {
     	if (!check_state(ucl, buf, DC_USER_SEND_GET))
@@ -753,10 +822,11 @@ user_input_available(DCUserConnLocal *ucl)
     int c;
     int res;
 
+    alarm(0); /* cannot fail */
     res = byteq_read(ucl->user_recvq, ucl->user_socket);
     if (res == 0 || (res < 0 && errno != EAGAIN)) {
  	warn_socket_error(res, false, _("user"));
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: socket error above */
 	return;
     }
 
@@ -788,23 +858,25 @@ user_input_available(DCUserConnLocal *ucl)
         byteq_remove(ucl->user_recvq, start);
 
     ucl->user_recvq_last = ucl->user_recvq->cur;
+    alarm(USER_CONN_IDLE_TIMEOUT); /* cannot fail */
 }
 
 static void
 user_now_writable(DCUserConnLocal *ucl)
 {
+    alarm(0); /* cannot fail */
     if (ucl->user_state == DC_USER_CONNECT) {
 	int error;
         socklen_t size = sizeof(error);
 
         if (getsockopt(ucl->user_socket, SOL_SOCKET, SO_ERROR, &error, &size) < 0) {
             warn(_("Cannot get error status - %s\n"), errstr);
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: local error */
 	    return;
         }
         if (error != 0) { /* The connect call on the socket failed. */
 	    warn(_("Cannot connect - %s\n"), strerror(error) /* not errno! */);
-	    ucl->user_running = false;
+	    terminate_process(ucl); /* MSG: local error */
 	    return;
 	}
 
@@ -837,8 +909,8 @@ user_now_writable(DCUserConnLocal *ucl)
 	    res = byteq_full_read_upto(ucl->user_sendq, ucl->transfer_fd/*UL*/, block);
 	    if (res < block) {
 	    	warn_file_error(res, false, ucl->local_file/*UL*/);
-	    	end_upload(ucl, false);
-		ucl->user_running = false;
+	    	end_upload(ucl, false, _("local error"));
+		terminate_process(ucl); /* MSG: local error */
 		return;
 	    }
     	    ucl->file_pos += res;
@@ -848,8 +920,8 @@ user_now_writable(DCUserConnLocal *ucl)
 	res = byteq_write(ucl->user_sendq, ucl->user_socket);
 	if (res == 0 || (res < 0 && errno != EAGAIN)) {
     	    warn_socket_error(res, true, _("user"));
-            end_upload(ucl, false);
-	    ucl->user_running = false;
+            end_upload(ucl, false, _("communication error"));
+	    terminate_process(ucl); /* MSG: communication error */
 	    return;
 	}
 	ucl->transfer_pos += res;
@@ -857,7 +929,7 @@ user_now_writable(DCUserConnLocal *ucl)
 
 	if (ucl->file_pos == ucl->file_size && ucl->user_sendq->cur == 0) {
     	    FD_CLR(ucl->user_socket, &ucl->user_write_fds);
-    	    end_upload(ucl, true);
+    	    end_upload(ucl, true, _("transfer complete"));
 	    ucl->user_state = DC_USER_GET;
 	}
     }
@@ -868,13 +940,14 @@ user_now_writable(DCUserConnLocal *ucl)
 	    res = byteq_write(ucl->user_sendq, ucl->user_socket);
 	    if (res == 0 || (res < 0 && errno != EAGAIN)) {
     	    	warn_socket_error(res, true, _("user"));
-	    	ucl->user_running = false;
+		terminate_process(ucl); /* MSG: socket error above */
 		return;
             }
 	}
 	if (ucl->user_sendq->cur == 0)
 	    FD_CLR(ucl->user_socket, &ucl->user_write_fds);
     }
+    alarm(USER_CONN_IDLE_TIMEOUT); /* cannot fail */
 }
 
 static void
@@ -886,11 +959,14 @@ signal_received(int signal)
     signal_char = signal;
     /* We don't care if this blocks - since we can't postpone this. */
     if (write(ucl->signal_pipe[1], &signal_char, sizeof(uint8_t)) < sizeof(uint8_t)) {
-	/* Die only if the signal is fatal. */
-	if (signal == SIGTERM || signal == SIGINT)
+	/* Die only if the signal is fatal.
+	 * SIGALRM is sent when the connection has been idle too long.
+	 * It will result in the connection shutting down, so if we
+	 * can't deliver it we might just shut down here.
+	 */
+	if (signal == SIGTERM || signal == SIGALRM)
 	    die(_("Cannot write to signal pipe - %s\n"), errstr); /* die OK */
-	else
-	    warn(_("Cannot write to signal pipe - %s\n"), errstr);
+        warn(_("Cannot write to signal pipe - %s\n"), errstr);
     }
 }
 
@@ -906,25 +982,56 @@ read_signal_input(DCUserConnLocal *ucl)
      */
     if (read(ucl->signal_pipe[0], &signal, sizeof(uint8_t)) < 0) {
 	warn(_("Cannot read from signal pipe - %s\n"), errstr);
-	ucl->user_running = false;
+	terminate_process(ucl); /* MSG: local error */
 	return;
     }
 
     if (signal == SIGTERM) {
-        ucl->user_running = false;
-    }
-    else if (signal == SIGUSR1) {
+        warn(_("Received TERM signal, shutting down.\n"));
+        terminate_process(ucl); /* MSG: terminating by signal */
+    } else if (signal == SIGALRM) {
+        warn(_("Idle timeout (%d seconds)\n"), USER_CONN_IDLE_TIMEOUT);
+	terminate_process(ucl); /* MSG: idle timeout msg above */
+    } else if (signal == SIGUSR1) {
         /* Not implemented yet, do nothing for now. */
     }
 }
 
+static void
+main_request_fd_writable(DCUserConnLocal *ucl)
+{
+    int res;
+
+    res = msgq_write(ucl->put_mq);
+    if (res <= 0) {
+        fatal_error(ucl, res, true);
+        return;
+    }
+    if (!msgq_has_partial_msg(ucl->put_mq))
+        FD_CLR(ucl->put_mq->fd, &ucl->user_write_fds);
+}
+
+static void
+main_result_fd_readable(DCUserConnLocal *ucl)
+{
+    int res;
+
+    res = msgq_read(ucl->get_mq);
+    if (res <= 0) {
+        fatal_error(ucl, res, false);
+        return;
+    }
+    warn(_("Received unknown message from main process, shutting down process.\n"));
+    ucl->user_running = false;
+}
+
 void
 __attribute__((noreturn))
-user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
+user_main(int get_fd[2], int put_fd[2], struct sockaddr_in *addr, int sock)
 {
     struct sigaction sigact;
     DCUserConnLocal *ucl;
-    
+
     ucl = xmalloc(sizeof(DCUserConnLocal));
     cur_ucl = ucl; /* one per process anyway */
     ucl->user_nick = NULL;
@@ -939,12 +1046,13 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
     ucl->file_size = 0;
     ucl->user_state = DC_USER_CONNECT;
     ucl->user_running = true;
-    ucl->main_socket = main_sockets[1];
-    ucl->ipc = ipc_new(ucl->main_socket, handle_main_msg, handle_main_ipc_error, ucl);
+    ucl->get_mq = msgq_new(get_fd[0]);
+    ucl->put_mq = msgq_new(put_fd[1]);
+    
     screen_writer = user_screen_writer;
 
-    if (close(main_sockets[0]) < 0)
-    	warn(_("Cannot close socket - %s\n"), errstr);
+    if (close(get_fd[1]) != 0 || close(put_fd[0]) != 0)
+        warn(_("Cannot close pipe - %s\n"), errstr);
 
     if (pipe(ucl->signal_pipe) < 0) {
     	warn(_("Cannot create pipe pair - %s\n"), errstr);
@@ -962,12 +1070,16 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
     /* Note: every signal registered with a non-ignore action in main.c
      * must also be registered here, either with an action or as ignored.
      */
-    if (sigaction(SIGTERM, &sigact, NULL) < 0 || sigaction(SIGUSR1, &sigact, NULL) < 0) {
+    if (sigaction(SIGTERM, &sigact, NULL) < 0
+            || sigaction(SIGUSR1, &sigact, NULL) < 0
+            || sigaction(SIGALRM, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
 	goto cleanup;
     }
     sigact.sa_handler = SIG_IGN;
-    if (sigaction(SIGINT, &sigact, NULL) < 0 || sigaction(SIGCHLD, &sigact, NULL) < 0) {
+    if (sigaction(SIGINT, &sigact, NULL) < 0
+          || sigaction(SIGCHLD, &sigact, NULL) < 0
+          || sigaction(SIGPIPE, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
 	goto cleanup;
     }
@@ -989,7 +1101,7 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 
     /* Set non-blocking I/O on socket. */
     if (!fd_set_nonblock_flag(ucl->user_socket, true)) {
-        warn(_("Cannot set socket flags - %s\n"), errstr);
+        warn(_("Cannot set non-blocking flag - %s\n"), errstr);
         goto cleanup;
     }
 
@@ -1008,8 +1120,8 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
     FD_ZERO(&ucl->user_read_fds);
     FD_ZERO(&ucl->user_write_fds);
     FD_SET(ucl->signal_pipe[0], &ucl->user_read_fds);
-    FD_SET(ucl->main_socket, &ucl->user_read_fds);
     FD_SET(ucl->user_socket, &ucl->user_write_fds); /* will set read after connect() finishes */
+    FD_SET(ucl->get_mq->fd, &ucl->user_read_fds);
 
     while (ucl->user_running) {
     	fd_set res_read_fds;
@@ -1028,8 +1140,10 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 
     	if (ucl->user_running && FD_ISSET(ucl->signal_pipe[0], &res_read_fds))
 	    read_signal_input(ucl);
-	if (ucl->user_running && FD_ISSET(ucl->main_socket, &res_read_fds))
-	    ipc_handle(ucl->ipc);
+        if (ucl->user_running && FD_ISSET(ucl->put_mq->fd, &res_write_fds))
+            main_request_fd_writable(ucl);
+        if (ucl->user_running && FD_ISSET(ucl->get_mq->fd, &res_read_fds))
+            main_result_fd_readable(ucl);
 	if (ucl->user_running && FD_ISSET(ucl->user_socket, &res_write_fds))
 	    user_now_writable(ucl);
 	if (ucl->user_running && FD_ISSET(ucl->user_socket, &res_read_fds))
@@ -1038,7 +1152,6 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 
 cleanup:
 
-    ipc_free(ucl->ipc);
     free(ucl->local_file);
     free(ucl->share_file);
     free(ucl->user_nick);
@@ -1051,10 +1164,15 @@ cleanup:
     	warn(_("Cannot close signal pipe - %s\n"), errstr);
     if (ucl->signal_pipe[1] >= 0 && close(ucl->signal_pipe[1]) < 0)
     	warn(_("Cannot close signal pipe - %s\n"), errstr);
-    if (ucl->main_socket >= 0 && close(ucl->main_socket) < 0)
-    	warn(_("Cannot close main process socket - %s\n"), errstr); /* XXX: "socket" or "connection" */
+    if (ucl->get_mq != NULL && close(ucl->get_mq->fd) != 0)
+    	warn(_("Cannot close pipe - %s\n"), errstr);
+    if (ucl->put_mq != NULL && close(ucl->put_mq->fd) != 0)
+    	warn(_("Cannot close pipe - %s\n"), errstr);
     if (ucl->user_socket >= 0 && close(ucl->user_socket) < 0)
     	warn(_("Cannot close user connection - %s\n"), errstr);
 
+    msgq_free(ucl->get_mq);
+    msgq_free(ucl->put_mq);
+
     exit(EXIT_SUCCESS);
 }
