@@ -25,6 +25,7 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <langinfo.h>		/* POSIX (XSI) */
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
@@ -40,7 +41,6 @@
 #include "common/quoting.h"
 #include "common/comparison.h"
 #include "common/bksearch.h"
-//#include "common/quoting.h"
 #include "microdc.h"
 
 /* TRANSLATORS: This comma-separated list specifies
@@ -58,6 +58,7 @@
  */
 #define POSITIVE_BOOL_STRINGS _("1,on,yes,true")
 
+static void var_set_charset(DCVariable *var, int argc, char **argv);
 static void var_set_nick(DCVariable *var, int argc, char **argv);
 static void var_set_password(DCVariable *var, int argc, char **argv);
 static void var_set_description(DCVariable *var, int argc, char **argv);
@@ -83,6 +84,7 @@ static void speed_completion_generator(DCCompletionInfo *ci);
 static void bool_completion_generator(DCCompletionInfo *ci);
 static void variable_completion_generator(DCCompletionInfo *ci);
 static void display_completion_generator(DCCompletionInfo *ci);
+static void charset_completion_generator(DCCompletionInfo *ci);
 
 typedef struct {
     DCDisplayFlag flag;
@@ -143,6 +145,13 @@ static DCVariable variables[] = {
       NULL,
       "Enable if listening for remote connections"
     },
+    {
+      "charset",
+      var_get_string, var_set_charset, &hub_charset,
+      charset_completion_generator,
+      NULL,
+      "Character set used for chat on the hub"
+    },
     {
       "description",
       var_get_string, var_set_description, &my_description,
@@ -260,6 +269,54 @@ find_variable(const char *name)
 		                   (comparison_fn_t) strcmp);
 }
 
+static void
+charset_completion_generator(DCCompletionInfo *ci)
+{
+    /* FIXME: NYI */
+    /*run the command  'iconv --list' to get a list of completion alternatives
+    fork();
+    exec('iconv', 'iconv', '--list');*/
+}
+
+static void
+var_set_charset(DCVariable *var, int argc, char **argv)
+{
+    iconv_t new_from_iconv = (iconv_t) -1;
+    iconv_t new_to_iconv = (iconv_t) -1;
+
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
+    }
+    if (argv[1][0] != '\0') {
+	char *system_charset;
+
+  	system_charset = nl_langinfo(CODESET);
+  	if (system_charset == NULL || *system_charset == '\0') {
+    	    warn(_("cannot get current character set"));
+	    return;
+	}
+        new_to_iconv = iconv_open(argv[1], system_charset);
+        if (new_to_iconv == (iconv_t) -1) {
+            warn(_("cannot create character set convertor - %s\n"), errstr);
+            return;
+        }
+        new_from_iconv = iconv_open(system_charset, argv[1]);
+        if (new_from_iconv == (iconv_t) -1) {
+            warn(_("cannot create character set convertor - %s\n"), errstr);
+            return;
+        }
+    }
+    if (to_hub_iconv != (iconv_t) -1 && iconv_close(to_hub_iconv) != 0)
+        warn(_("cannot close character convertor - %s\n"), errstr);
+    if (from_hub_iconv != (iconv_t) -1 && iconv_close(from_hub_iconv) != 0)
+        warn(_("cannot close character convertor - %s\n"), errstr);
+    to_hub_iconv = new_to_iconv;
+    from_hub_iconv = new_from_iconv;
+    free(hub_charset);
+    hub_charset = argv[1][0] == '\0' ? NULL : xstrdup(argv[1]);
+}
+
 static void
 var_set_nick(DCVariable *var, int argc, char **argv)
 {
@@ -521,7 +578,9 @@ var_set_listen_addr(DCVariable *var, int argc, char **argv)
     }
 
     if (!inet_aton(argv[1], &addr.sin_addr)) {
-	struct hostent *he;
+        screen_putf(_("%s: Specify listen address as an IP address\n"), quotearg(argv[1]));
+        /* XXX: fix this in the future... */
+	/*struct hostent *he;
 
 	screen_putf(_("Looking up IP address for %s\n"), quotearg(argv[1]));
 	he = gethostbyname(argv[1]);
@@ -530,7 +589,7 @@ var_set_listen_addr(DCVariable *var, int argc, char **argv)
 	    return;
 	}
 
-	addr.sin_addr = *(struct in_addr *) he->h_addr;
+	addr.sin_addr = *(struct in_addr *) he->h_addr;*/
     }
 
     force_listen_addr = addr.sin_addr;
